import pytest

from lionagi.core.mail import Mail
from lionagi.core.mail.package import Package
from datetime import datetime

"""
More to be added
"""


def trim_timestamp_to_day(timestamp_str):
    dt = datetime.fromisoformat(timestamp_str)
    return dt.strftime('%Y-%m-%d')


class MockPackage(Package):
    """A mock package class for testing purposes."""


@pytest.fixture
def mail():
    """Fixture to create a Mail instance with a MockPackage."""
    package = MockPackage()
    mail_instance = Mail(timestamp=datetime.today().strftime('%Y-%m-%d'), package=package)
    return mail_instance


def test_mail_category(mail):
    """Test the category property of Mail."""
    assert mail.category is None





# from lionagi.core.generic.mail import *
#
# import unittest
#
#
# class TestMail(unittest.TestCase):
#     def setUp(self):
#         self.mail = Mail(
#             sender="node1",
#             recipient="node2",
#             category=MailPackageCategory.MESSAGES,
#             package="Hello, World!"
#         )
#
#     def test_mail_initialization(self):
#         """Test initialization of Mail objects."""
#         self.assertIsInstance(self.mail, BaseComponent)
#         self.assertEqual(self.mail.sender, "node1")
#         self.assertEqual(self.mail.recipient, "node2")
#         self.assertEqual(self.mail.category, MailPackageCategory.MESSAGES)
#         self.assertEqual(self.mail.package, "Hello, World!")
#
#     def test_mail_str(self):
#         """Test the string representation of Mail."""
#         expected_str = "Mail from node1 to node2 with category messages and package Hello, World!"
#         self.assertEqual(str(self.mail), expected_str)
#
# class TestMailBox(unittest.TestCase):
#     def setUp(self):
#         self.mailbox = MailBox()
#         self.mail1 = Mail(
#             sender="node1",
#             recipient="node3",
#             category="model",
#             package={"model": "Random Forest"}
#         )
#         self.mail2 = Mail(
#             sender="node2",
#             recipient="node3",
#             category=MailPackageCategory.SERVICE,
#             package={"service": "Prediction"}
#         )
#
#     def test_adding_mails(self):
#         """Test adding mails to MailBox."""
#         self.mailbox.pending_ins["node1"] = self.mail1
#         self.mailbox.pending_outs["node3"] = self.mail2
#
#         self.assertIn("node1", self.mailbox.pending_ins)
#         self.assertIn("node3", self.mailbox.pending_outs)
#         self.assertEqual(self.mailbox.pending_ins["node1"], self.mail1)
#         self.assertEqual(self.mailbox.pending_outs["node3"], self.mail2)
#
#     def test_mailbox_str(self):
#         """Test the string representation of MailBox."""
#         self.mailbox.pending_ins["node1"] = self.mail1
#         self.mailbox.pending_outs["node3"] = self.mail2
#         expected_str = "MailBox with 1 pending incoming mails and 1 pending outgoing mails."
#         self.assertEqual(str(self.mailbox), expected_str)
#
# if __name__ == "__main__":
#     unittest.main()



#TODO
# import unittest
# from lionagi.core.generic.structure import *
#
# class TestCondition(Condition):
#     def check(self, node: Node) -> bool:
#         return True
#
# class TestBaseStructure(unittest.TestCase):
#     def setUp(self):
#         self.structure = BaseStructure(id_="test_structure")
#         self.node1 = Node(id_="node1", content="Node 1 content")
#         self.node2 = Node(id_="node2", content="Node 2 content")
#         self.node3 = Node(id_="node3", content="Node 3 content")
#
#     def test_internal_edges(self):
#         self.structure.add_node([self.node1, self.node2])
#         self.structure.relate_nodes(self.node1, self.node2)
#         self.assertEqual(len(self.structure.internal_edges), 1)
#
#     def test_is_empty(self):
#         self.assertTrue(self.structure.is_empty)
#         self.structure.add_node(self.node1)
#         self.assertFalse(self.structure.is_empty)
#
#     def test_add_node_single(self):
#         self.structure.add_node(self.node1)
#         self.assertIn(self.node1.id_, self.structure.internal_nodes)
#
#     def test_add_node_list(self):
#         self.structure.add_node([self.node1, self.node2])
#         self.assertIn(self.node1.id_, self.structure.internal_nodes)
#         self.assertIn(self.node2.id_, self.structure.internal_nodes)
#
#     def test_add_node_dict(self):
#         self.structure.add_node({self.node1.id_: self.node1, self.node2.id_: self.node2})
#         self.assertIn(self.node1.id_, self.structure.internal_nodes)
#         self.assertIn(self.node2.id_, self.structure.internal_nodes)
#
#     def test_add_node_duplicate(self):
#         self.structure.add_node(self.node1)
#         with self.assertRaises(ValueError):
#             self.structure.add_node(self.node1)
#
#     def test_get_node_str(self):
#         self.structure.add_node(self.node1)
#         self.assertEqual(self.structure.get_node(self.node1.id_), self.node1)
#
#     def test_get_node_node(self):
#         self.structure.add_node(self.node1)
#         self.assertEqual(self.structure.get_node(self.node1), self.node1)
#
#     def test_get_node_list(self):
#         self.structure.add_node([self.node1, self.node2])
#         self.assertEqual(self.structure.get_node([self.node1.id_, self.node2.id_]), [self.node1, self.node2])
#
#     def test_get_node_dict(self):
#         self.structure.add_node([self.node1, self.node2])
#         self.assertEqual(self.structure.get_node({self.node1.id_: self.node1, self.node2.id_: self.node2}), [self.node1, self.node2])
#
#     def test_get_node_not_found(self):
#         with self.assertRaises(KeyError):
#             self.structure.get_node("nonexistent_node")
#
#     def test_get_node_default(self):
#         self.assertIsNone(self.structure.get_node("nonexistent_node", default=None))
#
#     def test_remove_node_node(self):
#         self.structure.add_node(self.node1)
#         self.structure.remove_node(self.node1)
#         self.assertNotIn(self.node1.id_, self.structure.internal_nodes)
#
#     def test_remove_node_str(self):
#         self.structure.add_node(self.node1)
#         self.structure.remove_node(self.node1.id_)
#         self.assertNotIn(self.node1.id_, self.structure.internal_nodes)
#
#     def test_remove_node_list(self):
#         self.structure.add_node([self.node1, self.node2])
#         self.structure.remove_node([self.node1, self.node2])
#         self.assertNotIn(self.node1.id_, self.structure.internal_nodes)
#         self.assertNotIn(self.node2.id_, self.structure.internal_nodes)
#
#     def test_remove_node_dict(self):
#         self.structure.add_node([self.node1, self.node2])
#         self.structure.remove_node({self.node1.id_: self.node1, self.node2.id_: self.node2})
#         self.assertNotIn(self.node1.id_, self.structure.internal_nodes)
#         self.assertNotIn(self.node2.id_, self.structure.internal_nodes)
#
#     def test_pop_node_node(self):
#         self.structure.add_node(self.node1)
#         popped_node = self.structure.pop_node(self.node1)
#         self.assertEqual(popped_node, self.node1)
#         self.assertNotIn(self.node1.id_, self.structure.internal_nodes)
#
#     def test_pop_node_str(self):
#         self.structure.add_node(self.node1)
#         popped_node = self.structure.pop_node(self.node1.id_)
#         self.assertEqual(popped_node, self.node1)
#         self.assertNotIn(self.node1.id_, self.structure.internal_nodes)
#
#     def test_pop_node_list(self):
#         self.structure.add_node([self.node1, self.node2])
#         popped_nodes = self.structure.pop_node([self.node1, self.node2])
#         self.assertEqual(popped_nodes, [self.node1, self.node2])
#         self.assertNotIn(self.node1.id_, self.structure.internal_nodes)
#         self.assertNotIn(self.node2.id_, self.structure.internal_nodes)
#
#     def test_pop_node_dict(self):
#         self.structure.add_node([self.node1, self.node2])
#         popped_nodes = self.structure.pop_node({self.node1.id_: self.node1, self.node2.id_: self.node2})
#         self.assertEqual(popped_nodes, [self.node1, self.node2])
#         self.assertNotIn(self.node1.id_, self.structure.internal_nodes)
#         self.assertNotIn(self.node2.id_, self.structure.internal_nodes)
#
#     def test_pop_node_default(self):
#         self.assertIsNone(self.structure.pop_node("nonexistent_node", default=None))
#
#     def test_remove_edge_edge(self):
#         self.structure.add_node([self.node1, self.node2])
#         self.structure.relate_nodes(self.node1, self.node2)
#         edge = list(self.node1.edges.values())[0]
#         self.structure.remove_edge(edge)
#         self.assertNotIn(edge.id_, self.structure.internal_edges)
#
#     def test_remove_edge_str(self):
#         self.structure.add_node([self.node1, self.node2])
#         self.structure.relate_nodes(self.node1, self.node2)
#         edge = list(self.node1.edges.values())[0]
#         self.structure.remove_edge(edge.id_)
#         self.assertNotIn(edge.id_, self.structure.internal_edges)
#
#     def test_remove_edge_list(self):
#         self.structure.add_node([self.node1, self.node2, self.node3])
#         self.structure.relate_nodes(self.node1, self.node2)
#         self.structure.relate_nodes(self.node2, self.node3)
#         edges = list(self.node2.edges.values())
#
#         self.structure.remove_edge(edges)
#         self.assertNotIn(edges[0].id_, self.structure.internal_edges)
#         self.assertNotIn(edges[1].id_, self.structure.internal_edges)
#
#     def test_remove_edge_dict(self):
#         self.structure.add_node([self.node1, self.node2, self.node3])
#         self.structure.relate_nodes(self.node1, self.node2)
#         self.structure.relate_nodes(self.node2, self.node3)
#
#         edge_dict = self.node2.edges
#         edge_list = list(edge_dict.values())
#
#         self.structure.remove_edge(edge_dict)
#         self.assertNotIn(edge_list[0].id_, self.structure.internal_edges)
#         self.assertNotIn(edge_list[1].id_, self.structure.internal_edges)
#
#     def test_remove_edge_not_found(self):
#         with self.assertRaises(ValueError):
#             self.structure.remove_edge("nonexistent_edge")
#
#     def test_clear(self):
#         self.structure.add_node([self.node1, self.node2])
#         self.structure.clear()
#         self.assertTrue(self.structure.is_empty)
#
#     def test_get_node_edges_head(self):
#         self.structure.add_node([self.node1, self.node2])
#         self.structure.relate_nodes(self.node1, self.node2)
#         edges = self.structure.get_node_edges(self.node1, node_as="head")
#         self.assertEqual(len(edges), 1)
#
#     def test_get_node_edges_tail(self):
#         self.structure.add_node([self.node1, self.node2])
#         self.structure.relate_nodes(self.node1, self.node2)
#         edges = self.structure.get_node_edges(self.node2, node_as="tail")
#         self.assertEqual(len(edges), 1)
#
#     def test_get_node_edges_label(self):
#         self.structure.add_node([self.node1, self.node2])
#         self.structure.relate_nodes(self.node1, self.node2, label="test_label")
#         edges = self.structure.get_node_edges(self.node1, node_as="head", label="test_label")
#         self.assertEqual(len(edges), 1)
#
#     def test_add_edge(self):
#         self.structure.relate_nodes(self.node1, self.node2)
#         self.assertIn(self.node1.id_, self.structure.internal_nodes)
#         self.assertIn(self.node2.id_, self.structure.internal_nodes)
#         self.assertEqual(len(self.structure.internal_edges), 1)
#
#     def test_add_edge_with_label(self):
#         self.structure.relate_nodes(self.node1, self.node2, label="test_label")
#         edge = list(self.structure.internal_edges.values())[0]
#         self.assertEqual(edge.label, "test_label")
#
# if __name__ == "__main__":
#     unittest.main()


"""
Copyright 2024 HaiyangLi

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import ast
import operator
from typing import Any, Dict, Tuple, Callable

from lionagi.libs.ln_convert import to_dict


class BaseEvaluator:
    """
    A class to evaluate mathematical and boolean expressions from strings using Python's AST.

    Attributes:
        allowed_operators (Dict[type, Any]): A dictionary mapping AST node types to their corresponding Python operator functions.
        cache (Dict[Tuple[str, Tuple], Any]): A dictionary used to cache the results of evaluated expressions and sub-expressions.
    """

    def __init__(self) -> None:
        """Initializes the evaluator with supported operators and an empty cache."""
        self.allowed_operators: Dict[type, Any] = {
            ast.Add: operator.add,
            ast.Sub: operator.sub,
            ast.Mult: operator.mul,
            ast.Div: operator.truediv,
            ast.Pow: operator.pow,
            ast.Mod: operator.mod,
            ast.Eq: operator.eq,
            ast.NotEq: operator.ne,
            ast.Lt: operator.lt,
            ast.LtE: operator.le,
            ast.Gt: operator.gt,
            ast.GtE: operator.ge,
            ast.And: lambda x, y: x and y,
            ast.Or: lambda x, y: x or y,
            ast.Not: operator.not_,
            ast.USub: operator.neg,
        }
        self.cache: Dict[Tuple[str, Tuple], Any] = {}

    def evaluate(self, expression: str, context: Dict[str, Any]) -> Any:
        """
        Evaluates a given expression string using the provided context.

        Args:
            expression (str): The expression to evaluate.
            context (Dict[str, Any]): A dictionary mapping variable names to their values.

        Returns:
            Any: The result of the evaluated expression.

        Raises:
            ValueError: If the expression cannot be evaluated.
        """
        cache_key = (expression, tuple(sorted(context.items())))
        if cache_key in self.cache:
            return self.cache[cache_key]

        try:
            tree = ast.parse(expression, mode="eval")
            result = self._evaluate_node(tree.body, context)
            self.cache[cache_key] = result
            return result
        except Exception as e:
            raise ValueError(f"Failed to evaluate expression: {expression}. Error: {e}")

    def _evaluate_node(self, node: ast.AST, context: Dict[str, Any]) -> Any:
        """Recursively evaluates an AST node."""
        if isinstance(node, ast.BinOp):
            left = self._evaluate_node(node.left, context)
            op_func = self.allowed_operators[type(node.op)]
            right = self._evaluate_node(node.right, context)
            result = op_func(left, right)
        elif isinstance(node, ast.UnaryOp):
            operand = self._evaluate_node(node.operand, context)
            result = self.allowed_operators[type(node.op)](operand)
        elif isinstance(node, ast.Name):
            result = context.get(node.id, None)
        elif isinstance(node, ast.Constant):
            result = node.value
        elif isinstance(node, ast.Compare):
            left = self._evaluate_node(node.left, context)
            result = True
            for operation, comparator in zip(node.ops, node.comparators):
                op_func = self.allowed_operators[type(operation)]
                right = self._evaluate_node(comparator, context)
                result = result and op_func(left, right)
                if not result:
                    break
                left = right
        elif isinstance(node, ast.BoolOp):
            values = [self._evaluate_node(value, context) for value in node.values]
            if isinstance(node.op, ast.And):
                result = all(values)
            elif isinstance(node.op, ast.Or):
                result = any(values)
            else:
                raise ValueError("Unsupported boolean operation.")
        else:
            raise ValueError("Unsupported operation in condition.")
        return result

    def add_custom_operator(self, operator_name, operation_func):
        """Adds a custom operator to the evaluator."""
        custom_node_class = type(operator_name, (ast.AST,), {})
        if custom_node_class not in self.allowed_operators:
            self.allowed_operators[custom_node_class] = operation_func
        else:
            raise ValueError(f"Custom operator '{operator_name}' is already defined.")

    def evaluate_file(self, file_path, context, format="line"):
        """Evaluates expressions from a file."""
        if format == "line":
            with open(file_path, "r") as file:
                last_result = None
                for line in file:
                    line = line.strip()
                    if line:
                        last_result = self.evaluate(line, context)
                return last_result
        elif format == "json":
            with open(file_path, "r") as file:
                data = to_dict(file)
                last_result = None
                for expression in data:
                    last_result = self.evaluate(expression, context)
                return last_result
        else:
            raise ValueError(f"Unsupported file format: {format}")

    def validate_expression(self, expression):
        """Validates the given expression."""
        try:
            tree = ast.parse(expression, mode="eval")
            self._validate_node(tree.body)
            return True, "Expression is valid."
        except Exception as e:
            return False, f"Invalid expression: {str(e)}"

    def _validate_node(self, node):
        """Validates an AST node."""
        if isinstance(
            node, (ast.BinOp, ast.Compare, ast.BoolOp, ast.Name, ast.Constant)
        ):
            if (
                isinstance(node, ast.BinOp)
                and type(node.op) not in self.allowed_operators
            ):
                raise ValueError(
                    f"Operation {type(node.op).__name__} is not supported."
                )
        else:
            raise ValueError("Unsupported node type in expression.")


class BaseEvaluationEngine:
    def __init__(self) -> None:
        self.variables: Dict[str, Any] = {}
        self.functions: Dict[str, Callable] = {
            "print": print,
        }

    def _evaluate_expression(self, expression: str) -> Any:
        try:
            return eval(expression, {}, self.variables)
        except NameError as e:
            raise ValueError(f"Undefined variable. {e}")

    def _assign_variable(self, var_name: str, value: Any) -> None:
        self.variables[var_name] = value

    def _execute_function(self, func_name: str, *args: Any) -> None:
        if func_name in self.functions:
            self.functions[func_name](*args)
        else:
            raise ValueError(f"Function {func_name} not defined.")

    def _execute_statement(self, stmt: ast.AST) -> None:
        if isinstance(stmt, ast.Assign):
            var_name = stmt.targets[0].id
            value = self._evaluate_expression(ast.unparse(stmt.value))
            self._assign_variable(var_name, value)
        elif isinstance(stmt, ast.Expr) and isinstance(stmt.value, ast.Call):
            func_name = stmt.value.func.id
            args = [
                self._evaluate_expression(ast.unparse(arg)) for arg in stmt.value.args
            ]
            self._execute_function(func_name, *args)
        elif isinstance(stmt, ast.For):
            iter_var = stmt.target.id
            if isinstance(stmt.iter, ast.Call) and stmt.iter.func.id == "range":
                start, end = [
                    self._evaluate_expression(ast.unparse(arg))
                    for arg in stmt.iter.args
                ]
                for i in range(start, end):
                    self.variables[iter_var] = i
                    for body_stmt in stmt.body:
                        self._execute_statement(body_stmt)

    def execute(self, script: str) -> None:
        tree = ast.parse(script)
        for stmt in tree.body:
            self._execute_statement(stmt)



"""
Copyright 2024 HaiyangLi

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from lionagi.core.collections.abc import Field
from lionagi.core.report.form import Form


class BaseUnitForm(Form):
    """
    A base form class for units that includes fields for confidence scoring and reasoning.

    Attributes:
        template_name (str): The name of the template.
        confidence_score (float): A numeric confidence score between 0 and 1 with precision to 2 decimal places.
        reason (str | None): A field for providing concise reasoning for the process.
    """

    template_name: str = "UnitDirective"

    confidence_score: float = Field(
        None,
        description=(
            "Provide a numeric confidence score on how likely you successfully achieved the task  between 0 and 1, with precision in 2 decimal places. 1 being very confident in a good job, 0 is not confident at all."
        ),
        validation_kwargs={
            "upper_bound": 1,
            "lower_bound": 0,
            "num_type": float,
            "precision": 2,
        },
    )

    reason: str | None = Field(
        None,
        description=(
            "Explain yourself, provide concise reasoning for the process. Must start with: Let's think step by step, "
        ),
    )



"""
Copyright 2024 HaiyangLi

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from os import getenv
from lionagi.integrations.config.oai_configs import oai_schema
from lionagi.libs.ln_api import BaseService, PayloadPackage


allowed_kwargs = [
    "model",
    "frequency_penalty",
    "n",
    "presence_penalty",
    "response_format",
    "temperature",
    "top_p",
    "seed",
    "stop",
    "stream",
    "stream_options",
    "tools",
    "tool_choice",
    "user",
    "max_tokens",
    "logprobs",
    "top_logprobs",
    "logit_bias",
]


class OpenAIService(BaseService):
    """
    A service to interact with OpenAI's API endpoints.

    Attributes:
            base_url (str): The base URL for the OpenAI API.
            available_endpoints (list): A list of available API endpoints.
            schema (dict): The schema configuration for the API.
            key_scheme (str): The environment variable name for OpenAI API key.
            token_encoding_name (str): The default token encoding scheme.

    Examples:
            >>> service = OpenAIService(api_key="your_api_key")
            >>> asyncio.run(service.serve("Hello, world!","chat/completions"))
            (payload, completion)

            >>> service = OpenAIService()
            >>> asyncio.run(service.serve("Convert this text to speech.","audio_speech"))
    """

    base_url = "https://api.openai.com/v1/"
    available_endpoints = [
        "chat/completions",
        "finetune",
        "audio_speech",
        "audio_transcriptions",
        "audio_translations",
        "embeddings",
    ]
    schema = oai_schema
    key_scheme = "OPENAI_API_KEY"
    token_encoding_name = "cl100k_base"

    def __init__(
        self,
        api_key=None,
        key_scheme=None,
        schema=None,
        token_encoding_name: str = "cl100k_base",
        **kwargs,
    ):
        key_scheme = key_scheme or self.key_scheme
        super().__init__(
            api_key=api_key or getenv(key_scheme),
            schema=schema or self.schema,
            token_encoding_name=token_encoding_name,
            **kwargs,
        )
        self.active_endpoint = []
        self.allowed_kwargs = allowed_kwargs

    async def serve(self, input_, endpoint="chat/completions", method="post", **kwargs):
        """
        Serves the input using the specified endpoint and method.

        Args:
                input_: The input text to be processed.
                endpoint: The API endpoint to use for processing.
                method: The HTTP method to use for the request.
                **kwargs: Additional keyword arguments to pass to the payload creation.

        Returns:
                A tuple containing the payload and the completion assistant_response from the API.

        Raises:
                ValueError: If the specified endpoint is not supported.

        Examples:
                >>> service = OpenAIService(api_key="your_api_key")
                >>> asyncio.run(service.serve("Hello, world!","chat/completions"))
                (payload, completion)

                >>> service = OpenAIService()
                >>> asyncio.run(service.serve("Convert this text to speech.","audio_speech"))
                ValueError: 'audio_speech' is currently not supported
        """
        if endpoint not in self.active_endpoint:
            await self.init_endpoint(endpoint)
        if endpoint == "chat/completions":
            return await self.serve_chat(input_, **kwargs)
        else:
            return ValueError(f"{endpoint} is currently not supported")

    async def serve_chat(self, messages, required_tokens=None, **kwargs):
        """
        Serves the chat completion request with the given messages.

        Args:
                messages: The messages to be included in the chat completion.
                **kwargs: Additional keyword arguments for payload creation.

        Returns:
                A tuple containing the payload and the completion assistant_response from the API.

        Raises:
                Exception: If the API call fails.
        """
        if "chat/completions" not in self.active_endpoint:
            await self.init_endpoint("chat/completions")
            self.active_endpoint.append("chat/completions")

        msgs = []

        for msg in messages:
            if isinstance(msg, dict):
                content = msg.get("content")
                if isinstance(content, (dict, str)):
                    msgs.append({"role": msg["role"], "content": content})
                elif isinstance(content, list):
                    _content = []
                    for i in content:
                        if "text" in i:
                            _content.append({"type": "text", "text": str(i["text"])})
                        elif "image_url" in i:
                            _content.append(
                                {
                                    "type": "image_url",
                                    "image_url": {
                                        "url": f"{i['image_url'].get('url')}",
                                        "detail": i["image_url"].get("detail", "low"),
                                    },
                                }
                            )
                    msgs.append({"role": msg["role"], "content": _content})

        payload = PayloadPackage.chat_completion(
            msgs,
            self.endpoints["chat/completions"].config,
            self.schema["chat/completions"],
            **kwargs,
        )
        try:
            completion = await self.call_api(
                payload, "chat/completions", "post", required_tokens=required_tokens
            )
            return payload, completion
        except Exception as e:
            self.status_tracker.num_tasks_failed += 1
            raise e

    async def serve_embedding(self, embed_str, required_tokens=None, **kwargs):
        if "embeddings" not in self.active_endpoint:
            await self.init_endpoint("embeddings")
            self.active_endpoint.append("embeddings")

        payload = PayloadPackage.embeddings(
            embed_str,
            self.endpoints["embeddings"].config,
            self.schema["embeddings"],
            **kwargs,
        )

        try:
            embed = await self.call_api(
                payload, "embeddings", "post", required_tokens=required_tokens
            )
            return payload, embed
        except Exception as e:
            self.status_tracker.num_tasks_failed += 1
            raise e



import unittest
from lionagi.core.collections.abc import ItemNotFoundError
from lionagi.core.collections import Progression
from lionagi import Node


class TestProgression(unittest.TestCase):

    def setUp(self):
        self.nodes = [Node() for _ in range(5)]
        self.p = Progression(order=[node.ln_id for node in self.nodes])

    def test_initial_order(self):
        self.assertEqual(self.p.order, [node.ln_id for node in self.nodes])

    def test_append_node(self):
        new_node = Node()
        self.p.append(new_node)
        self.assertIn(new_node.ln_id, self.p.order)
        self.assertEqual(len(self.p), 6)

    def test_extend_single_node(self):
        new_node = Node()
        self.p.extend(new_node)
        self.assertIn(new_node.ln_id, self.p.order)
        self.assertEqual(len(self.p), 6)

    def test_extend_multiple_nodes(self):
        new_nodes = [Node() for _ in range(3)]
        self.p.extend(new_nodes)
        for node in new_nodes:
            self.assertIn(node.ln_id, self.p.order)
        self.assertEqual(len(self.p), 8)

    def test_include_node(self):
        new_node = Node()
        self.p.include(new_node)
        self.assertIn(new_node.ln_id, self.p.order)

    def test_getitem_slice(self):
        slice_result = self.p[1:5]
        self.assertEqual(slice_result.order, self.p.order[1:5])

    def test_setitem(self):
        new_node = Node()
        self.p[0] = new_node.ln_id
        self.assertEqual(self.p[0], new_node.ln_id)

    def test_add_node(self):
        new_node = Node()
        new_prog = self.p + new_node
        self.assertIn(new_node.ln_id, new_prog.order)

    def test_iadd_node(self):
        new_node = Node()
        self.p += new_node
        self.assertIn(new_node.ln_id, self.p.order)

    def test_subtract_node(self):
        node_to_remove = self.nodes[0]
        self.p -= node_to_remove.ln_id
        self.assertNotIn(node_to_remove.ln_id, self.p.order)

    def test_isubtract_node(self):
        node_to_remove = self.nodes[0]
        self.p -= node_to_remove.ln_id
        self.assertNotIn(node_to_remove.ln_id, self.p.order)

    def test_popleft(self):
        leftmost = self.p.order[0]
        self.assertEqual(self.p.popleft(), leftmost)
        self.assertNotIn(leftmost, self.p.order)

    def test_exclude_node(self):
        node_to_exclude = self.nodes[0]
        self.p.exclude(node_to_exclude.ln_id)
        self.assertNotIn(node_to_exclude.ln_id, self.p.order)

    def test_clear(self):
        self.p.clear()
        self.assertEqual(len(self.p), 0)

    def test_to_dict(self):
        p_dict = self.p.to_dict()
        self.assertEqual(p_dict["order"], self.p.order)
        self.assertEqual(p_dict["name"], self.p.name)

    def test_bool(self):
        self.assertTrue(bool(self.p))

    def test_remove_nonexistent_node(self):
        non_existent_node = Node()
        with self.assertRaises(ItemNotFoundError):
            self.p.remove(non_existent_node.ln_id)

    def test_pop_index_error(self):
        with self.assertRaises(ItemNotFoundError):
            self.p.pop(100)

    def test_popleft_index_error(self):
        empty_prog = Progression()
        with self.assertRaises(ItemNotFoundError):
            empty_prog.popleft()

    def test_iteration(self):
        nodes_list = [node.ln_id for node in self.nodes]
        for i, node in enumerate(self.p):
            self.assertEqual(node, nodes_list[i])

    def test_contains(self):
        node_to_check = self.nodes[0]
        self.assertIn(node_to_check.ln_id, self.p)

    def test_not_contains(self):
        non_existent_node = Node()
        self.assertNotIn(non_existent_node.ln_id, self.p)

    def test_equality(self):
        new_prog = Progression(order=[node.ln_id for node in self.nodes])
        self.assertEqual(self.p.order, new_prog.order)

    def test_inequality(self):
        new_node = Node()
        new_prog = self.p + new_node
        self.assertNotEqual(self.p.order, new_prog)


if __name__ == "__main__":
    unittest.main()



import unittest
from lionagi.core.report.form import Form
from lionagi.core.report.report import Report


class TestForm(unittest.TestCase):

    def setUp(self):
        self.form = Form(assignment="input1, input2 -> output")

    def test_initial_fields(self):
        self.assertEqual(self.form.input_fields, ["input1", "input2"])
        self.assertEqual(self.form.requested_fields, ["output"])
        self.assertFalse(self.form.workable)
        self.assertFalse(self.form.filled)
        self.assertEqual(
            self.form.work_fields, {"input1": None, "input2": None, "output": None}
        )

    def test_fill_input_fields(self):
        self.form.fill(input1=1, input2=2)
        self.assertTrue(self.form.workable)
        self.assertFalse(self.form.filled)
        self.assertEqual(
            self.form.work_fields, {"input1": 1, "input2": 2, "output": None}
        )

    def test_fill_output_field(self):
        self.form.fill(input1=1, input2=2)
        self.form.fill(output=3)
        self.assertTrue(self.form.filled)
        self.assertEqual(self.form.work_fields, {"input1": 1, "input2": 2, "output": 3})
        self.assertFalse(self.form.workable)

    def test_fill_again_raises_error(self):
        self.form.fill(input1=1, input2=2, output=3)
        with self.assertRaises(ValueError):
            self.form.fill(input1=2, input2=3)


class TestReport(unittest.TestCase):

    def setUp(self):
        self.report = Report(assignment="a, b -> c")

    def test_initial_fields(self):
        self.assertEqual(self.report.input_fields, ["a", "b"])
        self.assertEqual(self.report.requested_fields, ["c"])
        self.assertEqual(self.report.work_fields, {"a": None, "b": None, "c": None})
        self.assertFalse(self.report.filled)
        self.assertFalse(self.report.workable)

    def test_fill_input_fields(self):
        self.report.fill(a=3, b=4)
        self.assertEqual(self.report.work_fields, {"a": 3, "b": 4, "c": None})
        self.assertTrue(self.report.workable)

    def test_fill_output_field(self):
        self.report.fill(a=3, b=4)
        self.report.fill(c=4)
        self.assertEqual(self.report.work_fields, {"a": 3, "b": 4, "c": 4})
        self.assertTrue(self.report.filled)
        self.assertFalse(self.report.workable)

    def test_next_forms_none(self):
        self.assertIsNone(self.report.next_forms())

    def test_complex_assignment(self):
        self.report = Report(
            assignment="a, b -> h",
            form_assignments=[
                "a, b -> c",
                "a -> e",
                "b -> f",
                "c -> g",
                "e, f, g -> h",
            ],
            a=3,
            b=4,
        )
        self.assertEqual(
            self.report.work_fields,
            {
                "a": None,
                "b": None,
                "c": None,
                "e": None,
                "f": None,
                "g": None,
                "h": None,
            },
        )
        self.report.fill(c=5, e=6, f=7, g=8, h=10)
        self.assertEqual(
            self.report.work_fields,
            {"a": None, "b": None, "c": 5, "e": 6, "f": 7, "g": 8, "h": 10},
        )
        self.assertFalse(self.report.filled)
        self.assertFalse(self.report.workable)
        with self.assertRaises(ValueError):
            self.report.fill(c="xx", e="yy")


if __name__ == "__main__":
    unittest.main()



import unittest
from lionagi import Node, pile, progression
from lionagi.core.collections.abc import Element
from lionagi.core.collections import Exchange


class TestExchange(unittest.TestCase):

    def setUp(self):
        # Setup for creating initial nodes, piles, and progressions
        self.nodes = [Node(content=i) for i in range(10)]
        self.pile1 = pile(self.nodes)
        self.prog1 = progression(self.nodes[:5], name="left")
        self.prog2 = progression(self.nodes[5:], name="right")
        self.exchange = Exchange()

    def test_initialization(self):
        self.assertEqual(len(self.exchange.pile), 0)
        self.assertEqual(len(self.exchange.pending_ins), 0)
        self.assertEqual(len(self.exchange.pending_outs), 0)

    def test_include_in(self):
        sender_id = "sender1"
        for node in self.nodes[:5]:
            node.sender = sender_id
            self.exchange.include(node, "in")

        self.assertEqual(len(self.exchange.pile), 5)
        self.assertIn(sender_id, self.exchange.pending_ins)
        self.assertEqual(len(self.exchange.pending_ins[sender_id]), 5)

    def test_include_out(self):
        for node in self.nodes[:5]:
            self.exchange.include(node, "out")

        self.assertEqual(len(self.exchange.pile), 5)
        self.assertEqual(len(self.exchange.pending_outs), 5)

    def test_exclude(self):
        for node in self.nodes[:5]:
            self.exchange.include(node, "out")

        node_to_remove = self.nodes[0]
        self.exchange.exclude(node_to_remove)
        self.assertNotIn(node_to_remove, self.exchange.pile)
        self.assertNotIn(node_to_remove, self.exchange.pending_outs)

    def test_exclude_from_in(self):
        sender_id = "sender1"
        for node in self.nodes[:5]:
            node.sender = sender_id
            self.exchange.include(node, "in")

        node_to_remove = self.nodes[0]
        self.exchange.exclude(node_to_remove)
        self.assertNotIn(node_to_remove, self.exchange.pile)
        self.assertNotIn(node_to_remove, self.exchange.pending_ins[sender_id])

    def test_senders(self):
        sender_id = "sender1"
        for node in self.nodes[:5]:
            node.sender = sender_id
            self.exchange.include(node, "in")

        self.assertIn(sender_id, self.exchange.senders)

    def test_to_dict(self):
        for node in self.nodes[:5]:
            self.exchange.include(node, "out")

        exchange_dict = self.exchange.to_dict()
        self.assertIn("pile", exchange_dict)
        self.assertIn("pending_ins", exchange_dict)
        self.assertIn("pending_outs", exchange_dict)

    def test_bool(self):
        self.assertTrue(self.exchange)

    # Additional Tests for Edge Cases
    def test_include_non_sendable(self):
        non_sendable = Element()
        with self.assertRaises(AttributeError):
            self.exchange.include(non_sendable, "in")

    def test_exclude_nonexistent_item(self):
        non_existent_node = Node(content="nonexistent")
        result = self.exchange.exclude(non_existent_node)
        self.assertTrue(result)

    def test_include_multiple_items_in(self):
        sender_id = "sender2"
        new_nodes = [Node(content="new_node1"), Node(content="new_node2")]
        for new_node in new_nodes:
            new_node.sender = sender_id
            self.exchange.include(new_node, "in")

        self.assertEqual(len(self.exchange.pile), 2)
        self.assertEqual(len(self.exchange.pending_ins[sender_id]), 2)

    def test_include_multiple_items_out(self):
        new_nodes = [Node(content="new_node1"), Node(content="new_node2")]
        for new_node in new_nodes:
            self.exchange.include(new_node, "out")

        self.assertEqual(len(self.exchange.pile), 2)
        self.assertEqual(len(self.exchange.pending_outs), 2)

    def test_exclude_from_empty_exchange(self):
        result = self.exchange.exclude(Node(content="nonexistent"))
        self.assertTrue(result)

    def test_clear_empty_exchange(self):
        self.exchange.pile.clear()
        self.exchange.pending_ins.clear()
        self.exchange.pending_outs.clear()
        self.assertEqual(len(self.exchange.pile), 0)
        self.assertEqual(len(self.exchange.pending_ins), 0)
        self.assertEqual(len(self.exchange.pending_outs), 0)

    def test_clear_exchange(self):
        sender_id = "sender3"
        for node in self.nodes[:5]:
            node.sender = sender_id
            self.exchange.include(node, "in")
        for node in self.nodes[5:]:
            self.exchange.include(node, "out")

        self.exchange.pile.clear()
        self.exchange.pending_ins.clear()
        self.exchange.pending_outs.clear()

        self.assertEqual(len(self.exchange.pile), 0)
        self.assertEqual(len(self.exchange.pending_ins), 0)
        self.assertEqual(len(self.exchange.pending_outs), 0)


if __name__ == "__main__":
    unittest.main()



"""
Copyright 2024 HaiyangLi

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from enum import Enum
from .base import BaseUnitForm, Field


class SelectTemplate(BaseUnitForm):
    """
    A template for performing a selection task based on given instructions and context.

    Inherits from `BaseUnitForm` and adds fields and methods specific to selection tasks.

    Attributes:
        confidence_score (float | None): A numeric score between 0 to 1 formatted to 2 decimal places, indicating confidence in the selection.
        reason (str | None): A brief reason for the given output.
        template_name (str): The name of the template.
        selection (Enum | str | list | None): The selection from given choices.
        choices (list): The given choices to select from.
        assignment (str): A string representing the task assignment for the selection.
    """

    confidence_score: float | None = Field(
        None,
        description="a numeric score between 0 to 1 formatted in num:0.2f, 1 being very confident and 0 being not confident at all, just guessing",
        validation_kwargs={
            "upper_bound": 1,
            "lower_bound": 0,
            "num_type": float,
            "precision": 2,
        },
    )

    reason: str | None = Field(
        default_factory=str,
        description="brief reason for the given output, format: This is my best response because ...",
    )

    template_name: str = "default_select"

    selection: Enum | str | list | None = Field(
        None, description="selection from given choices"
    )
    choices: list = Field(default_factory=list, description="the given choices")

    assignment: str = "task -> selection"

    @property
    def answer(self):
        """
        Gets the selection attribute.

        Returns:
            Enum | str | list | None: The selection from given choices.
        """
        return getattr(self, "selection", None)

    def __init__(
        self,
        *,
        instruction=None,
        context=None,
        choices=None,
        reason=False,
        confidence_score=False,
        **kwargs,
    ):
        """
        Initializes a new instance of the SelectTemplate class.

        Args:
            instruction (str, optional): Additional instructions for the selection task.
            context (str, optional): Additional context for the selection task.
            choices (list, optional): The choices to select from.
            reason (bool, optional): Whether to include a reasoning field. Defaults to False.
            confidence_score (bool, optional): Whether to include a confidence score. Defaults to False.
            **kwargs: Additional keyword arguments.
        """
        super().__init__(**kwargs)

        self.choices = choices
        self.task = f"""
select 1 item from the provided choices {choices}.        
1. additional objective: {instruction or "N/A"}.
2. additional information: {context or "N/A"}.     
"""
        if reason:
            self.append_to_request("reason")

        if confidence_score:
            self.append_to_request("confidence_score")



"""
Copyright 2024 HaiyangLi

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from lionagi.libs.ln_convert import strip_lower
from lionagi.libs.ln_func_call import rcall
from lionagi.core.collections.abc import Directive
from lionagi.core.validator.validator import Validator
from lionagi.core.collections import iModel
from .unit_form import UnitForm
from .unit_mixin import DirectiveMixin
from .util import retry_kwargs


class Unit(Directive, DirectiveMixin):
    """
    Unit is a class that extends Directive and DirectiveMixin to provide
    advanced operations like chat, direct actions, and predictions using a
    specific branch and model.

    Attributes:
        branch (Branch): The branch instance associated with the Unit.
        imodel (iModel): The model instance used for the Unit.
        form_template (Type[Form]): The form template to use for operations.
        validator (Validator): The validator instance for response validation.
    """

    default_template = UnitForm

    def __init__(
        self, branch, imodel: iModel = None, template=None, rulebook=None, verbose=False
    ) -> None:
        self.branch = branch
        if imodel and isinstance(imodel, iModel):
            branch.imodel = imodel
            self.imodel = imodel
        else:
            self.imodel = branch.imodel
        self.form_template = template or self.default_template
        self.validator = Validator(rulebook=rulebook) if rulebook else Validator()
        self.verbose = verbose

    async def chat(
        self,
        instruction=None,
        context=None,
        system=None,
        sender=None,
        recipient=None,
        branch=None,
        requested_fields=None,
        form=None,
        tools=False,
        invoke_tool=True,
        return_form=True,
        strict=False,
        rulebook=None,
        imodel=None,
        clear_messages=False,
        use_annotation=True,
        return_branch=False,
        **kwargs,
    ):
        """
        Asynchronously performs a chat operation.

        Args:
            instruction (str, optional): Instruction message.
            context (str, optional): Context message.
            system (str, optional): System message.
            sender (str, optional): Sender identifier.
            recipient (str, optional): Recipient identifier.
            branch (Branch, optional): Branch instance.
            requested_fields (list, optional): Fields requested in the response.
            form (Form, optional): Form data.
            tools (bool, optional): Flag indicating if tools should be used.
            invoke_tool (bool, optional): Flag indicating if tools should be invoked.
            return_form (bool, optional): Flag indicating if form should be returned.
            strict (bool, optional): Flag indicating if strict validation should be applied.
            rulebook (Rulebook, optional): Rulebook instance for validation.
            imodel (iModel, optional): Model instance.
            clear_messages (bool, optional): Flag indicating if messages should be cleared.
            use_annotation (bool, optional): Flag indicating if annotations should be used.
            return_branch (bool, optional): Flag indicating if branch should be returned.
            kwargs: Additional keyword arguments.

        Returns:
            Any: The processed response.
        """
        kwargs = {**retry_kwargs, **kwargs}
        return await rcall(
            self._chat,
            instruction=instruction,
            context=context,
            system=system,
            sender=sender,
            recipient=recipient,
            branch=branch,
            requested_fields=requested_fields,
            form=form,
            tools=tools,
            invoke_tool=invoke_tool,
            return_form=return_form,
            strict=strict,
            rulebook=rulebook,
            imodel=imodel,
            clear_messages=clear_messages,
            use_annotation=use_annotation,
            return_branch=return_branch,
            **kwargs,
        )

    async def direct(
        self,
        instruction=None,
        *,
        context=None,
        form=None,
        branch=None,
        tools=None,
        return_branch=False,
        reason: bool = False,
        predict: bool = False,
        score=None,
        select=None,
        plan=None,
        allow_action: bool = False,
        allow_extension: bool = False,
        max_extension: int = None,
        confidence=None,
        score_num_digits=None,
        score_range=None,
        select_choices=None,
        plan_num_step=None,
        predict_num_sentences=None,
        directive: str = None,
        verbose=None,
        **kwargs,
    ):
        """
        Asynchronously directs the operation based on the provided parameters.

        Args:
            instruction (str, optional): Instruction message.
            context (str, optional): Context message.
            form (Form, optional): Form data.
            branch (Branch, optional): Branch instance.
            tools (Any, optional): Tools to be used.
            return_branch (bool, optional): Flag indicating if branch should be returned.
            reason (bool, optional): Flag indicating if reason should be included.
            predict (bool, optional): Flag indicating if prediction should be included.
            score (Any, optional): Score parameters.
            select (Any, optional): Select parameters.
            plan (Any, optional): Plan parameters.
            allow_action (bool, optional): Flag indicating if action should be allowed.
            allow_extension (bool, optional): Flag indicating if extension should be allowed.
            max_extension (int, optional): Maximum extension value.
            confidence (Any, optional): Confidence parameters.
            score_num_digits (int, optional): Number of digits for score.
            score_range (tuple, optional): Range for score.
            select_choices (list, optional): Choices for selection.
            plan_num_step (int, optional): Number of steps for plan.
            predict_num_sentences (int, optional): Number of sentences for prediction.
            directive (str, optional): Directive for the operation.
            kwargs: Additional keyword arguments.

        Returns:
            Any: The processed response.
        """
        kwargs = {**retry_kwargs, **kwargs}
        verbose = verbose if verbose is not None else self.verbose

        if not directive:

            out = await rcall(
                self._direct,
                instruction=instruction,
                context=context,
                form=form,
                branch=branch,
                tools=tools,
                return_branch=return_branch,
                reason=reason,
                predict=predict,
                score=score,
                select=select,
                plan=plan,
                allow_action=allow_action,
                allow_extension=allow_extension,
                max_extension=max_extension,
                confidence=confidence,
                score_num_digits=score_num_digits,
                score_range=score_range,
                select_choices=select_choices,
                plan_num_step=plan_num_step,
                predict_num_sentences=predict_num_sentences,
                verbose=verbose,
                **kwargs,
            )

            if verbose:
                print(
                    "\n--------------------------------------------------------------"
                )
                print(f"Directive successfully completed!")

            return out

        out = await rcall(
            self._mono_direct,
            directive=directive,
            instruction=instruction,
            context=context,
            branch=branch,
            tools=tools,
            verbose=verbose,
            **kwargs,
        )

        if verbose:
            print("--------------------------------------------------------------")
            print(f"Directive successfully completed!")

        return out

    async def select(self, *args, **kwargs):
        """
        Asynchronously performs a select operation using the _select method with
        retry logic.

        Args:
            *args: Positional arguments to pass to the _select method.
            **kwargs: Keyword arguments to pass to the _select method, including
                retry configurations.

        Returns:
            Any: The result of the select operation.
        """
        from .template.select import SelectTemplate

        kwargs = {**retry_kwargs, **kwargs}
        kwargs["template"] = kwargs.get("template", SelectTemplate)
        return await rcall(self._select, *args, **kwargs)

    async def predict(self, *args, **kwargs):
        """
        Asynchronously performs a predict operation using the _predict method with
        retry logic.

        Args:
            *args: Positional arguments to pass to the _predict method.
            **kwargs: Keyword arguments to pass to the _predict method, including
                retry configurations.

        Returns:
            Any: The result of the predict operation.
        """
        from .template.predict import PredictTemplate

        kwargs = {**retry_kwargs, **kwargs}
        kwargs["template"] = kwargs.get("template", PredictTemplate)
        return await rcall(self._predict, *args, **kwargs)

    async def score(self, *args, **kwargs):
        """
        Asynchronously performs a score operation using the _score method with retry logic.

        Args:
            *args: Positional arguments to pass to the _score method.
            **kwargs: Keyword arguments to pass to the _score method, including retry configurations.

        Returns:
            Any: The result of the score operation.
        """
        from .template.score import ScoreTemplate

        kwargs = {**retry_kwargs, **kwargs}
        kwargs["template"] = kwargs.get("template", ScoreTemplate)
        return await rcall(self._score, *args, **kwargs)

    async def plan(self, *args, **kwargs):
        """
        Asynchronously performs a plan operation using the _plan method with retry logic.

        Args:
            *args: Positional arguments to pass to the _plan method.
            **kwargs: Keyword arguments to pass to the _plan method, including retry configurations.

        Returns:
            Any: The result of the plan operation.
        """
        from .template.plan import PlanTemplate

        kwargs = {**retry_kwargs, **kwargs}
        kwargs["template"] = kwargs.get("template", PlanTemplate)
        return await rcall(self._plan, *args, **kwargs)

    async def _mono_direct(
        self,
        directive: str,  # examples, "chat", "predict", "act"
        instruction=None,  # additional instruction
        context=None,  # context to perform the instruction on
        system=None,  # optionally swap system message
        branch=None,
        tools=None,
        template=None,
        verbose=None,
        **kwargs,
    ):
        """
        Asynchronously performs a single direct operation.

        Args:
            directive (str): The directive for the operation.
            instruction (str, optional): Additional instruction.
            context (str, optional): Context for the operation.
            system (str, optional): System message.
            branch (Branch, optional): Branch instance.
            tools (Any, optional): Tools to be used.
            template (Any, optional): Template for the operation.
            kwargs: Additional keyword arguments.

        Returns:
            Any: The result of the direct operation.
        """

        if template:
            kwargs["template"] = template

        kwargs = {**retry_kwargs, **kwargs}
        branch = branch or self.branch

        if system:
            branch.add_message(system=system)

        if hasattr(self, strip_lower(directive)):
            directive = getattr(self, strip_lower(directive))

            verbose = verbose if verbose is not None else self.verbose
            if verbose:
                print(f"Performing directive: {directive}...")

            return await directive(
                context=context,
                instruction=instruction,
                tools=tools,
                **kwargs,
            )

        raise ValueError(f"invalid directive: {directive}")



"""
Copyright 2024 HaiyangLi

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from lionagi.libs.ln_func_call import rcall, pcall
from lionagi.libs import convert, AsyncUtil

from lionagi.core.collections.abc import Directive
from lionagi.core.collections import iModel
from lionagi.core.validator.validator import Validator
from lionagi.core.session.branch import Branch
from lionagi.core.unit.util import retry_kwargs


class ParallelUnit(Directive):
    """
    A class representing a unit that can perform parallel chat interactions.

    Attributes:
        branch (Session): The session to which this ParallelUnit belongs.
        imodel (iModel): The model to be used for interactions.
        form_template (Form): The template for the form to be used.
        validator (Validator): The validator to validate the forms.

    Methods:
        pchat(*args, **kwargs): Asynchronously performs a parallel chat interaction.
        _parallel_chat(
            instruction,
            num_instances=1,
            context=None,
            sender=None,
            messages=None,
            tools=False,
            out=True,
            invoke=True,
            requested_fields=None,
            persist_path=None,
            branch_config={},
            explode=False,
            include_mapping=True,
            default_key="response",
            **kwargs,
        ): Asynchronously performs the core logic for parallel chat interactions.
    """

    default_template = None

    def __init__(
        self, session, imodel: iModel = None, template=None, rulebook=None
    ) -> None:
        """
        Initializes a new instance of ParallelUnit.

        Args:
            session (Session): The session to which this ParallelUnit belongs.
            imodel (iModel, optional): The model to be used for interactions.
            template (Form, optional): The template for the form to be used.
            rulebook (Rulebook, optional): The rulebook to validate the forms.
        """
        self.branch = session
        if imodel and isinstance(imodel, iModel):
            session.imodel = imodel
            self.imodel = imodel
        else:
            self.imodel = session.imodel
        self.form_template = template or self.default_template
        self.validator = Validator(rulebook=rulebook) if rulebook else Validator()

    async def pchat(self, *args, **kwargs):
        """
        Asynchronously performs a parallel chat interaction.

        Args:
            *args: Positional arguments to pass to the _parallel_chat method.
            **kwargs: Keyword arguments to pass to the _parallel_chat method,
                      including retry configurations.

        Returns:
            Any: The result of the parallel chat interaction.
        """
        kwargs = {**retry_kwargs, **kwargs}
        return await rcall(self._parallel_chat, *args, **kwargs)

    async def _parallel_chat(
        self,
        instruction: str,
        num_instances=1,
        context=None,
        sender=None,
        messages=None,
        tools=False,
        out=True,
        invoke: bool = True,
        requested_fields=None,
        persist_path=None,
        branch_config={},
        explode=False,
        include_mapping=True,
        default_key="response",
        **kwargs,
    ):
        """
        Asynchronously performs the core logic for parallel chat interactions.

        Args:
            instruction (str): The instruction to perform.
            num_instances (int, optional): Number of instances to run in parallel.
                                           Defaults to 1.
            context (Any, optional): The context to perform the instruction on.
            sender (str, optional): The sender of the instruction. Defaults to None.
            messages (list, optional): The list of messages. Defaults to None.
            tools (bool, optional): Flag indicating if tools should be used.
                                    Defaults to False.
            out (bool, optional): Flag indicating if output should be returned.
                                  Defaults to True.
            invoke (bool, optional): Flag indicating if tools should be invoked.
                                     Defaults to True.
            requested_fields (list, optional): Fields to request from the context.
                                               Defaults to None.
            persist_path (str, optional): Path to persist the branch. Defaults to None.
            branch_config (dict, optional): Configuration for the branch. Defaults to {}.
            explode (bool, optional): Flag indicating if combinations of instructions
                                      and context should be exploded. Defaults to False.
            include_mapping (bool, optional): Flag indicating if instruction, context,
                                              and branch mapping should be included.
                                              Defaults to True.
            default_key (str, optional): The default key for the response. Defaults to "response".
            **kwargs: Additional keyword arguments for further customization.

        Returns:
            Any: The result of the parallel chat interaction.
        """

        branches = {}

        async def _inner(i, ins_, cxt_):

            branch_ = Branch(
                messages=messages,
                service=self.session.default_branch.service,
                llmconfig=self.session.default_branch.llmconfig,
                persist_path=persist_path,
                **branch_config,
            )

            branch_.branch_name = branch_.id_

            if tools:
                branch_.tool_manager = self.session.default_branch.tool_manager

            res_ = await branch_.chat(
                instruction=ins_ or instruction,
                context=cxt_ or context,
                sender=sender,
                tools=tools,
                invoke=invoke,
                out=out,
                requested_fields=requested_fields,
                **kwargs,
            )

            branches[branch_.id_] = branch_
            if include_mapping:
                return {
                    "instruction": ins_ or instruction,
                    "context": cxt_ or context,
                    "branch_id": branch_.id_,
                    default_key: res_,
                }

            else:
                return res_

        async def _inner_2(i, ins_=None, cxt_=None):
            """returns num_instances of branches performing for same task/context"""
            tasks = [_inner(i, ins_, cxt_) for _ in range(num_instances)]
            ress = await AsyncUtil.execute_tasks(*tasks)
            return convert.to_list(ress)

        async def _inner_3(i):
            """different instructions but same context"""
            tasks = [_inner_2(i, ins_=ins_) for ins_ in convert.to_list(instruction)]
            ress = await AsyncUtil.execute_tasks(*tasks)
            return convert.to_list(ress)

        async def _inner_3_b(i):
            """different context but same instruction"""
            tasks = [_inner_2(i, cxt_=cxt_) for cxt_ in convert.to_list(context)]
            ress = await AsyncUtil.execute_tasks(*tasks)
            return convert.to_list(ress)

        async def _inner_4(i):
            """different instructions and different context"""

            tasks = []
            if explode:
                tasks = [
                    _inner_2(i, ins_=ins_, cxt_=cxt_)
                    for ins_ in convert.to_list(instruction)
                    for cxt_ in convert.to_list(context)
                ]
            else:
                tasks = [
                    _inner_2(i, ins_=ins_, cxt_=cxt_)
                    for ins_, cxt_ in zip(
                        convert.to_list(instruction), convert.to_list(context)
                    )
                ]

            ress = await AsyncUtil.execute_tasks(*tasks)
            return convert.to_list(ress)

        if len(convert.to_list(instruction)) == 1:
            if len(convert.to_list(context)) == 1:
                out_ = await _inner_2(0)
                self.session.branches.update(branches)
                return out_

            elif len(convert.to_list(context)) > 1:
                out_ = await _inner_3_b(0)
                self.session.branches.update(branches)
                return out_

        elif len(convert.to_list(instruction)) > 1:
            if len(convert.to_list(context)) == 1:
                out_ = await _inner_3(0)
                self.session.branches.update(branches)
                return out_

            elif len(convert.to_list(context)) > 1:
                out_ = await _inner_4(0)
                self.session.branches.update(branches)
                return out_



"""
Copyright 2024 HaiyangLi

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from pydantic import Field

from lionagi.core.collections.abc import Actionable
from lionagi.core.generic.node import Node
from .tool import Tool


class DirectiveSelection(Node, Actionable):
    """
    Represents a directive selection node which can be invoked to perform an action.

    Attributes:
        directive (str): The action to be performed, with a default value of "chat".
        directive_kwargs (dict): The arguments for the action.

    Methods:
        invoke(): An asynchronous method to perform the action defined by the directive.
    """

    directive: str = Field(
        "chat", description="The action to be performed", alias="action_type"
    )
    directive_kwargs: dict = Field(
        default_factory=dict,
        description="The arguments for the action",
        alias="action_arguments",
    )

    async def invoke(self):
        """
        Perform the action defined by the directive.

        This method is intended to be overridden by subclasses to provide specific
        implementation details for the action.
        """
        pass


class ActionNode(DirectiveSelection):
    """
    Represents an action node that can invoke actions within a branch using tools and instructions.

    Attributes:
        tools (list[Tool] | Tool | None): The tools to be used in the action.
        instruction (Node): The instruction for the action.

    Methods:
        invoke(branch, context=None): An asynchronous method to invoke the action
                                      within the given branch.
    """

    tools: list[Tool] | Tool | None = Field(
        default_factory=list,
        description="The tools to be used in the action",
        alias="tool",
    )
    instruction: Node = Field(
        ..., description="The instruction for the action", alias="instruct"
    )

    async def invoke(self, branch, context=None):
        """
        Invoke the action within the given branch.

        Args:
            branch: The branch in which to perform the action.
            context: Optional; Additional context for the action.

        Returns:
            The result of the action, depending on the directive.

        Raises:
            ValueError: If the directive is not "chat" or "direct".
        """
        if self.directive == "chat":
            return await branch.chat(
                instruction=self.instruction.instruct,
                tools=self.tools,
                **self.directive_kwargs,
            )
        elif self.directive == "direct":
            if self.tools:
                self.directive_kwargs["allow_action"] = True
            return await branch.direct(
                instruction=self.instruction.instruct,
                context=context,
                tools=self.tools,
                **self.directive_kwargs,
            )
        else:
            raise ValueError(
                'Invalid directive, valid directives are: "chat", "direct"'
            )



"""
Copyright 2024 HaiyangLi

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import ast
from functools import lru_cache
from lionagi.libs import AsyncUtil
from ..evaluator.base_evaluator import BaseEvaluator
from .sandbox_ import SandboxTransformer


class ScriptEngine:
    def __init__(self):
        self.variables = {}
        self.safe_evaluator = BaseEvaluator()
        self.functions = {
            "processData": self.process_data,
        }

    def process_data(self, data):
        return data * 2

    def _evaluate_expression(self, expression):
        return self.safe_evaluator.evaluate(expression, self.variables)

    def _assign_variable(self, var_name, value):
        self.variables[var_name] = value

    def _execute_function(self, func_name, arg):
        if func_name in self.functions:
            return self.functions[func_name](arg)
        else:
            raise ValueError(f"Function {func_name} not defined.")

    def execute(self, script):
        tree = ast.parse(script)
        for stmt in tree.body:
            if isinstance(stmt, ast.Assign):
                var_name = stmt.targets[0].id
                value = self._evaluate_expression(ast.unparse(stmt.value))
                self._assign_variable(var_name, value)
            elif isinstance(stmt, ast.Expr) and isinstance(stmt.value, ast.Call):
                func_name = stmt.value.func.id
                arg = self._evaluate_expression(ast.unparse(stmt.value.args[0]))
                result = self._execute_function(func_name, arg)
                # For demonstration, manually update 'x' to simulate expected behavior
                if func_name == "processData":
                    self._assign_variable("x", result)

    def add_hook(self, event_name, callback):
        """Subscribe a callback function to a specific event."""
        self.hooks[event_name].append(callback)

    def trigger_hooks(self, event_name, *args, **kwargs):
        """Trigger all callbacks attached to a specific event."""
        for callback in self.hooks[event_name]:
            callback(*args, **kwargs)

    async def process_data(self, data):
        # Example asynchronous function
        return data * 2

    @lru_cache(maxsize=128)
    def evaluate_expression(self, expression):
        return self.safe_evaluator.evaluate(expression, self.variables)

    async def _execute_function_async(self, func_name, arg):
        if func_name in self.functions:
            func = self.functions[func_name]
            if AsyncUtil.is_coroutine_func(func):
                return await func(arg)
            else:
                return func(arg)
        else:
            raise ValueError(f"Function {func_name} not defined.")

    def execute_sandboxed(self, script):
        # Parse and sanitize the script
        tree = ast.parse(script, mode="exec")
        sanitized_tree = SandboxTransformer().visit(tree)
        ast.fix_missing_locations(sanitized_tree)

        # Compile the sanitized AST
        code = compile(sanitized_tree, "<sandbox>", "exec")

        # Execute the code in a restricted namespace
        exec(code, {"__builtins__": None}, self.variables)



#TODO
# import unittest
# from lionagi.new.schema.todo.tree import Tree
# from lionagi.core.generic import TreeNode
#
#
# class TestTree(unittest.TestCase):
#     def setUp(self):
#         self.tree = Tree()
#         self.node1 = TreeNode(id_="node1", content="Node 1")
#         self.node2 = TreeNode(id_="node2", content="Node 2")
#         self.node3 = TreeNode(id_="node3", content="Node 3")
#         self.tree.add_structure_node(self.node1)
#         self.tree.add_structure_node(self.node2)
#         self.tree.add_structure_node(self.node3)
#
#     def test_add_parent_to_child(self):
#         self.tree.add_parent_to_child(self.node1, self.node2)
#         self.assertEqual(self.node2, self.node1.parent)
#         self.assertIn(self.node1, self.node2.children)
#
#     def test_add_child_to_parent(self):
#         self.tree.add_child_to_parent(self.node3, self.node2)
#         self.assertEqual(self.node2, self.node3.parent)
#         self.assertIn(self.node3, self.node2.children)
#
#     def test_find_parent(self):
#         self.tree.add_parent_to_child(self.node1, self.node2)
#         parent = self.tree.find_parent(self.node1)
#         self.assertEqual(self.node2, parent)
#
#     def test_find_child(self):
#         self.tree.add_parent_to_child(self.node1, self.node2)
#         children = self.tree.find_child(self.node2)
#         self.assertIn(self.node1, children)
#
#     def test_parent_child_relationship(self):
#         self.tree.add_parent_to_child(self.node1, self.node2)
#         self.tree.add_child_to_parent(self.node3, self.node2)
#         self.assertIn(self.node1, self.node2.children)
#         self.assertIn(self.node3, self.node2.children)
#         self.assertEqual(self.node2, self.node1.parent)
#         self.assertEqual(self.node2, self.node3.parent)
#
#     # Add more tests as necessary to cover edge cases and other functionalities
#
#
# if __name__ == "__main__":
#     unittest.main()



import unittest
from lionagi.core.generic.tree_node import TreeNode
from lionagi.core.generic.tree import Tree


class TestTreeNode(unittest.TestCase):

    def setUp(self):
        self.tree_node1 = TreeNode()
        self.tree_node2 = TreeNode()
        self.tree_node3 = TreeNode()
        self.tree_node4 = TreeNode()
        self.tree = Tree()

    def test_initialization(self):
        self.assertIsNone(self.tree_node1.parent)
        self.assertEqual(self.tree_node1.children, [])
        self.assertIsInstance(self.tree_node1, TreeNode)
        self.assertIsInstance(self.tree_node1.ln_id, str)

    def test_relate_child(self):
        self.tree_node1.relate_child([self.tree_node2, self.tree_node3])
        self.assertIn(self.tree_node2.ln_id, self.tree_node1.children)
        self.assertIn(self.tree_node3.ln_id, self.tree_node1.children)
        self.assertEqual(self.tree_node2.parent, self.tree_node1)
        self.assertEqual(self.tree_node3.parent, self.tree_node1)

    def test_relate_parent(self):
        self.tree_node1.relate_parent(self.tree_node4)
        self.assertEqual(self.tree_node1.parent, self.tree_node4)
        self.assertIn(self.tree_node1.ln_id, self.tree_node4.children)

    def test_unrelate_parent(self):
        self.tree_node1.relate_parent(self.tree_node4)
        self.tree_node1.unrelate_parent()
        self.assertIsNone(self.tree_node1.parent)

    def test_unrelate_child(self):
        self.tree_node1.relate_child([self.tree_node2, self.tree_node3])
        self.tree_node1.unrelate_child([self.tree_node2, self.tree_node3])
        self.assertNotIn(self.tree_node2.ln_id, self.tree_node1.children)
        self.assertNotIn(self.tree_node3.ln_id, self.tree_node1.children)
        self.assertIsNone(self.tree_node2.parent)
        self.assertIsNone(self.tree_node3.parent)

    def test_tree_integration(self):
        self.tree.relate_parent_child(
            self.tree_node1, [self.tree_node2, self.tree_node3, self.tree_node4]
        )
        self.assertEqual(len(self.tree.internal_nodes), 4)
        self.assertEqual(len(self.tree_node1.children), 3)
        self.assertEqual(self.tree.root, self.tree_node1)

    def test_tree_node_edges(self):
        self.tree_node1.relate_child([self.tree_node2, self.tree_node3])
        df = self.tree_node1.edges.to_df()
        self.assertEqual(len(df), 2)
        self.assertIn(self.tree_node2.ln_id, df["tail"].values)
        self.assertIn(self.tree_node3.ln_id, df["tail"].values)

    def test_tree_display(self):
        # This test is mainly to check if the display function runs without errors
        self.tree.relate_parent_child(
            self.tree_node1, [self.tree_node2, self.tree_node3, self.tree_node4]
        )
        try:
            self.tree.display()
        except Exception as e:
            self.fail(f"Tree display method raised an exception: {e}")


if __name__ == "__main__":
    unittest.main()



import pytest
from api.apicore import _connect as _connect_
from api.apicore import _config as _config_


"""
Setup the testing construct to ease future testing cases
"""


@pytest.fixture()
def config():
    _config = _config_.ConfigSingleton()
    _config.config["MOCK"] = True  # turn on mock mode
    yield _config


@pytest.fixture
def test_s3_conn(config):
    yield _connect_.get_object("AWSS3")


@pytest.fixture
def test_ec2_conn(config):
    yield _connect_.get_object("AWSEC2")






from lionagi.libs.ln_convert import strip_lower


def get_input_output_fields(str_: str) -> list[list[str]]:
    """
    Parses an assignment string to extract input and output fields.

    Args:
        str_ (str): The assignment string in the format 'inputs -> outputs'.

    Returns:
        list[list[str]]: A list containing two lists - one for input fields and one for requested fields.

    Raises:
        ValueError: If the assignment string is None or if it does not contain '->' indicating invalid format.
    """
    if str_ is None:
        return [], []

    if "->" not in str_:
        raise ValueError("Invalid assignment format. Expected 'inputs -> outputs'.")

    inputs, outputs = str_.split("->")

    input_fields = [strip_lower(i) for i in inputs.split(",")]
    requested_fields = [strip_lower(o) for o in outputs.split(",")]

    return input_fields, requested_fields



"""
Copyright 2024 HaiyangLi

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

"""
This module defines the FunctionCalling class, which facilitates dynamic
invocation of functions based on various input types. It supports initializing
function calls from tuples, dictionaries, ActionRequest objects, or JSON strings.

Note:
    Function Calling object is the only way for AI system to call functions.
"""

from functools import singledispatchmethod
from typing import Any, Callable, Dict

from lionagi.libs import ParseUtil
from lionagi.libs.ln_func_call import call_handler
from lionagi.core.collections.abc import Actionable
from lionagi.core.message.action_request import ActionRequest


class FunctionCalling(Actionable):
    """
    A class for dynamically invoking functions based on various input types,
    allowing for specification of the function and arguments through multiple
    formats including tuples, dictionaries, ActionRequests, or JSON strings.
    """

    def __init__(self, function: Callable, arguments: Dict[str, Any] = None):
        """
        Initializes a new instance of FunctionCalling with the given function
        and optional arguments.

        Args:
            function (Callable): The function to be called.
            arguments (Dict[str, Any]): Arguments to pass to the function.
                Defaults to an empty dictionary.
        """
        self.function = function
        self.arguments = arguments or {}

    @property
    def func_name(self) -> str:
        """
        Returns the name of the function.

        Returns:
            str: The function's name.
        """
        return self.function.__name__

    @singledispatchmethod
    @classmethod
    def create(cls, func_call: Any) -> "FunctionCalling":
        """
        Creates an instance of FunctionCalling based on the type of input.

        Args:
            func_call (Any): The function call description, which can be a tuple, dict,
                ActionRequest, or JSON string.

        Returns:
            FunctionCalling: An instance of FunctionCalling prepared to invoke
                the specified function.

        Raises:
            TypeError: If the input type is not supported.
        """
        raise TypeError(f"Unsupported type {type(func_call)}")

    @create.register(tuple)
    def _(cls, function_calling: tuple) -> "FunctionCalling":
        if len(function_calling) == 2:
            return cls(function=function_calling[0], arguments=function_calling[1])
        else:
            raise ValueError(f"Invalid function call {function_calling}")

    @create.register(dict)
    def _(cls, function_calling: Dict[str, Any]) -> "FunctionCalling":
        if len(function_calling) == 2 and (
            {"function", "arguments"} <= function_calling.keys()
        ):
            return cls.create(
                (function_calling["function"], function_calling["arguments"])
            )
        raise ValueError(f"Invalid function call {function_calling}")

    @create.register(ActionRequest)
    def _(cls, function_calling: ActionRequest) -> "FunctionCalling":
        return cls.create((function_calling.function, function_calling.arguments))

    @create.register(str)
    def _(cls, function_calling: str) -> "FunctionCalling":
        _call = None
        try:
            _call = ParseUtil.fuzzy_parse_json(function_calling)
        except Exception as e:
            raise ValueError(f"Invalid function call {function_calling}") from e

        if isinstance(_call, dict):
            return cls.create(_call)
        raise ValueError(f"Invalid function call {function_calling}")

    async def invoke(self) -> Any:
        """
        Asynchronously invokes the stored function with the provided arguments.

        Returns:
            Any: The result of the function call.
        """
        return await call_handler(self.function, **self.arguments)

    def __str__(self) -> str:
        """
        Returns a string representation of the function call.

        Returns:
            str: String representation of the function call.
        """
        return f"{self.func_name}({self.arguments})"

    def __repr__(self) -> str:
        return self.__str__()



#TODO
# import unittest
# from lionagi.core.graph.tree import TreeNode, Tree
#
# class TestTreeNode(unittest.TestCase):
#     def setUp(self):
#         self.parent_node = TreeNode(id_="parent", content="Parent Node")
#         self.child_node1 = TreeNode(id_="child1", content="Child Node 1")
#         self.child_node2 = TreeNode(id_="child2", content="Child Node 2")
#
#     def test_relate_child(self):
#         self.parent_node.relate_child(self.child_node1)
#         self.assertIn("child1", self.parent_node.children)
#         self.assertEqual(self.child_node1.parent, self.parent_node)
#
#     def test_relate_children(self):
#         self.parent_node.relate_child([self.child_node1, self.child_node2])
#         self.assertIn("child1", self.parent_node.children)
#         self.assertIn("child2", self.parent_node.children)
#         self.assertEqual(self.child_node1.parent, self.parent_node)
#         self.assertEqual(self.child_node2.parent, self.parent_node)
#
#     def test_relate_parent(self):
#         self.child_node1.relate_parent(self.parent_node)
#         self.assertIn("child1", self.parent_node.children)
#         self.assertEqual(self.child_node1.parent, self.parent_node)
#
#     def test_unrelate_child(self):
#         self.parent_node.relate_child(self.child_node1)
#         self.parent_node.unrelate_child(self.child_node1)
#         self.assertNotIn("child1", self.parent_node.children)
#         self.assertIsNone(self.child_node1.parent)
#
#     def test_unrelate_parent(self):
#         self.child_node1.relate_parent(self.parent_node)
#         self.child_node1.unrelate_parent()
#         self.assertNotIn("child1", self.parent_node.children)
#         self.assertIsNone(self.child_node1.parent)
#
# class TestTree(unittest.TestCase):
#     def setUp(self):
#         self.tree = Tree()
#         self.root = TreeNode(id_="root", content="Root Node")
#         self.child_node1 = TreeNode(id_="child1", content="Child Node 1")
#         self.child_node2 = TreeNode(id_="child2", content="Child Node 2")
#         self.grandchild_node = TreeNode(id_="grandchild", content="Grandchild Node")
#
#     def test_add_node(self):
#         self.tree.add_node(self.root)
#         self.assertIn("root", self.tree.internal_nodes)
#
#     def test_relate_parent_child(self):
#         self.tree.relate_parent_child(self.root, [self.child_node1, self.child_node2])
#         self.assertIn("child1", self.root.children)
#         self.assertIn("child2", self.root.children)
#         self.assertEqual(self.tree.root, self.root)
#
#     def test_tree_structure(self):
#         # Build the tree
#         self.tree.relate_parent_child(self.root, [self.child_node1, self.child_node2])
#         self.tree.relate_parent_child(self.child_node1, self.grandchild_node)
#
#         # Check the tree structure
#         self.assertIn("grandchild", self.child_node1.children)
#         self.assertEqual(self.grandchild_node.parent, self.child_node1)
#         self.assertEqual(self.child_node1.parent, self.root)
#
#     def test_acyclic(self):
#         # Build the tree
#         self.tree.relate_parent_child(self.root, self.child_node1)
#         self.tree.relate_parent_child(self.child_node1, self.child_node2)
#         # Trees should always be acyclic
#         self.assertTrue(self.tree.acyclic)
#
# if __name__ == "__main__":
#     unittest.main()



"""
Copyright 2024 HaiyangLi

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from lionagi.core.collections.abc import Progressable
from lionagi.core.collections import pile, progression, Pile
from lionagi.core.work.work import Work, WorkStatus
from lionagi.core.work.work_queue import WorkQueue


class WorkLog(Progressable):
    """
    A class representing a log of work items.

    Attributes:
        pile (Pile): A pile containing work items.
        pending (Progression): A progression of pending work items.
        queue (WorkQueue): A queue to manage the execution of work items.
    """

    def __init__(self, capacity=10, workpile=None, refresh_time=1):
        """
        Initializes a new instance of WorkLog.

        Args:
            capacity (int): The capacity of the work queue batch processing.
            workpile (Pile, optional): An optional pile of initial work items.
            refresh_time (int, optional): The time interval to refresh the work log queue.
                Defaults to 1.
        """
        self.pile = (
            workpile if workpile and isinstance(workpile, Pile) else pile({}, Work)
        )
        self.pending = progression(workpile) if workpile else progression()
        self.queue = WorkQueue(capacity=capacity, refresh_time=refresh_time)

    async def append(self, work: Work):
        """
        Appends a new work item to the log.

        Args:
            work (Work): The work item to append.
        """
        self.pile.append(work)
        self.pending.append(work)

    async def forward(self):
        """
        Forwards pending work items to the queue.
        """
        while len(self.pending) > 0:
            work: Work = self.pile[self.pending.popleft()]
            await self.queue.enqueue(work)

    async def stop(self):
        """
        Stops the work queue.
        """
        await self.queue.stop()

    @property
    def pending_work(self):
        """
        Retrieves the pile of pending work items.

        Returns:
            Pile: A pile of pending work items.
        """
        return pile([i for i in self.pile if i.status == WorkStatus.PENDING])

    @property
    def stopped(self):
        """
        Checks if the work queue is stopped.

        Returns:
            bool: True if the work queue is stopped, else False.
        """
        return self.queue.stopped

    @property
    def completed_work(self):
        """
        Retrieves the pile of completed work items.

        Returns:
            Pile: A pile of completed work items.
        """
        return pile([i for i in self.pile if i.status == WorkStatus.COMPLETED])

    def __contains__(self, work):
        """
        Checks if a work item is in the pile.

        Args:
            work (Work): The work item to check.

        Returns:
            bool: True if the work item is in the pile, else False.
        """
        return work in self.pile

    def __iter__(self):
        """
        Returns an iterator over the work pile.

        Returns:
            Iterator: An iterator over the work pile.
        """
        return iter(self.pile)



"""
Copyright 2024 HaiyangLi

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

"""
This module contains the BaseAgent class, which serves as a base class for agents.
"""

from pydantic import Field
from typing import Any, Callable

from lionagi.libs import func_call, AsyncUtil


from lionagi.core.mail.start_mail import StartMail
from lionagi.core.generic.node import Node
from lionagi.core.mail.mail_manager import MailManager
from lionagi.core.executor.base_executor import BaseExecutor
from lionagi.core.executor.graph_executor import GraphExecutor


class BaseAgent(Node):

    def __init__(
        self,
        structure: BaseExecutor,
        executable: BaseExecutor,
        output_parser=None,
        **kwargs,
    ) -> None:
        """
        Initializes the BaseAgent instance.

        Args:
            structure: The structure of the agent.
            executable_obj: The executable object of the agent.
            output_parser: A function for parsing the agent's output (optional).
        """
        super().__init__(**kwargs)
        self.structure: BaseExecutor = structure
        self.executable: BaseExecutor = executable
        self.start: StartMail = StartMail()
        self.mail_manager: MailManager = MailManager(
            [self.structure, self.executable, self.start]
        )
        self.output_parser: Callable | None = output_parser
        self.start_context: Any | None = None

    async def mail_manager_control(self, refresh_time=1):
        """
        Controls the mail manager execution based on the structure and executable states.

        Args:
            refresh_time: The time interval (in seconds) for checking the execution states (default: 1).
        """
        while not self.structure.execute_stop or not self.executable.execute_stop:
            await AsyncUtil.sleep(refresh_time)
        self.mail_manager.execute_stop = True

    async def execute(self, context=None):
        """
        Executes the agent with the given context and returns the parsed output (if available).

        Args:
            context: The initial context for the agent (optional).

        Returns:
            The parsed output of the agent (if available).
        """
        self.start_context = context
        self.start.trigger(
            context=context,
            structure_id=self.structure.ln_id,
            executable_id=self.executable.ln_id,
        )
        await func_call.mcall(
            [0.1, 0.1, 0.1, 0.1],
            [
                self.structure.execute,
                self.executable.execute,
                self.mail_manager.execute,
                self.mail_manager_control,
            ],
        )

        self.structure.execute_stop = False
        self.executable.execute_stop = False
        self.mail_manager.execute_stop = False

        if self.output_parser:
            return self.output_parser(self)



import unittest
from lionagi.core.collections.abc import LionTypeError, ItemNotFoundError
from lionagi.core.generic import Graph, Node, Edge
from lionagi.libs.ln_convert import to_list


class TestGraph(unittest.TestCase):

    def setUp(self):
        """Set up a basic graph for testing."""
        self.node1 = Node()
        self.node2 = Node()
        self.node3 = Node()
        self.node4 = Node()
        self.g = Graph()

    def test_initial_size(self):
        """Test the initial size of the graph."""
        self.assertEqual(self.g.size(), 0)

    def test_initial_internal_edges(self):
        """Test the initial number of internal edges."""
        self.assertEqual(len(self.g.internal_edges), 0)

    def test_add_node_no_edge(self):
        """Test that adding nodes does not create any edges."""
        self.g.add_node([self.node1, self.node2, self.node3, self.node4])
        self.assertEqual(len(self.g.internal_edges), 0)

    def test_add_edges(self):
        """Test adding edges between nodes."""
        self.g.add_edge(self.node1, self.node2)
        self.g.add_edge(self.node2, self.node3)
        self.g.add_edge(self.node3, self.node4)
        self.assertEqual(len(self.g.internal_edges), 3)

    def test_get_node(self):
        """Test retrieving a node by its identifier."""
        self.g.add_node(self.node2)
        node5 = self.g.get_node(self.node2)
        self.assertEqual(node5, self.node2)

    def test_get_node_edges(self):
        """Test getting the edges of a node."""
        self.g.add_edge(self.node1, self.node2)
        self.assertEqual(len(self.g.get_node_edges(self.node1)), 1)

    def test_get_heads(self):
        """Test getting head nodes."""
        self.g.add_edge(self.node1, self.node2)
        self.g.add_edge(self.node2, self.node3)
        self.g.add_edge(self.node3, self.node4)
        self.assertEqual(len(self.g.get_heads()), 1)

    def test_is_acyclic(self):
        """Test if the graph is acyclic."""
        self.g.add_edge(self.node1, self.node2)
        self.g.add_edge(self.node2, self.node3)
        self.g.add_edge(self.node3, self.node4)
        self.assertTrue(self.g.is_acyclic())

    def test_remove_edge(self):
        """Test removing an edge from the graph."""
        self.g.add_edge(self.node1, self.node2)
        self.g.add_edge(self.node2, self.node3)
        self.g.add_edge(self.node3, self.node4)
        edge = self.g.internal_edges[0]
        self.g.remove_edge(edge)
        self.assertEqual(len(self.g.internal_edges), 2)

    def test_remove_nonexistent_edge(self):
        """Test removing a nonexistent edge."""
        self.g.add_edge(self.node1, self.node2)
        with self.assertRaises(ItemNotFoundError):
            self.g.remove_edge(edge=Edge(head=self.node2, tail=self.node3))

    def test_clear(self):
        """Test clearing all nodes and edges from the graph."""
        self.g.add_node([self.node1, self.node2, self.node3, self.node4])
        self.g.add_edge(self.node1, self.node2)
        self.g.add_edge(self.node2, self.node3)
        self.g.add_edge(self.node3, self.node4)
        self.g.clear()
        self.assertTrue(self.g.is_empty())
        self.assertEqual(len(self.g.internal_edges), 0)

    def test_display(self):
        """Test displaying the graph (visual check)."""
        self.g.add_edge(self.node1, self.node2)
        self.g.add_edge(self.node2, self.node3)
        self.g.add_edge(self.node3, self.node4)
        self.g.display()  # This should display the graph visually


if __name__ == "__main__":
    unittest.main()



"""
Copyright 2024 HaiyangLi

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from lionagi.libs.ln_convert import to_str
from .base import BaseUnitForm, Field


class PlanTemplate(BaseUnitForm):
    """
    A template for generating a step-by-step plan based on given instructions and context.

    Inherits from `BaseUnitForm` and adds fields and methods specific to plan generation.

    Attributes:
        template_name (str): The name of the template.
        plan (dict | str): The generated step-by-step plan in the format {step_n: {plan: ..., reason: ...}}.
        signature (str): A string representing the task signature for the plan.
    """

    template_name: str = "plan_template"

    plan: dict | str = Field(
        default_factory=dict,
        description="the generated step by step plan, return as a dictionary following {step_n: {plan: ..., reason: ...}} format",
    )

    assignment: str = "task -> plan"

    @property
    def answer(self):
        """
        Gets the plan attribute.

        Returns:
            dict | str: The generated plan.
        """
        return getattr(self, "plan", None)

    def __init__(
        self,
        *,
        instruction=None,
        context=None,
        confidence_score=False,
        reason=False,
        num_step=3,
        **kwargs,
    ):
        """
        Initializes a new instance of the PlanTemplate class.

        Args:
            instruction (str, optional): Additional instructions for the plan.
            context (str, optional): Additional context for the plan.
            confidence_score (bool, optional): Whether to include a confidence score.
            reason (bool, optional): Whether to include a reasoning field.
            num_step (int, optional): The number of steps in the plan. Defaults to 3.
            **kwargs: Additional keyword arguments.
        """
        super().__init__(**kwargs)

        self.task = f"""
Generate a {num_step}_step plan based on the given context
1. additional instruction, {to_str(instruction or "N/A")}
2. additional context, {to_str(context or "N/A")}
"""
        if reason:
            self.append_to_request("reason")

        if confidence_score:
            self.append_to_request("confidence_score")



"""
Copyright 2024 HaiyangLi

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

"""This module defines abstract base classes for LionAGI."""

from abc import ABC, abstractmethod
from collections.abc import Generator
from typing import Any, Iterator, TypeVar

from pydantic import Field, BaseModel, field_validator

from .component import LionIDable, get_lion_id
from .exceptions import LionTypeError

T = TypeVar("T")


class Record(ABC):
    """
    Abstract base class for managing a collection of unique LionAGI items.

    Accepts LionIDable for retrieval and requires Component instances for
    addition.

    Methods:
        keys: Return an iterator over the ln_id of items in the record.
        values: Return an iterator over items in the record.
        get: Retrieve an item by identifier.
        __getitem__: Return an item using a LionIDable identifier.
        __setitem__: Add or update an item in the record.
        __contains__: Check if an item is in the record.
        __len__: Return the number of items in the record.
        __iter__: Iterate over items in the record.
    """

    @abstractmethod
    def keys(self) -> Generator[str, None, None]:
        """Return an iterator over the ln_id of items in the record."""

    @abstractmethod
    def values(self) -> Generator[T, None, None]:
        """Return an iterator over items in the record."""

    @abstractmethod
    def get(self, item: LionIDable, /, default: Any = None) -> T:
        """
        Retrieve an item by identifier.

        Accepts a LionIDable object. Returns the default if the item is not
        found.

        Args:
            item (LionIDable): The identifier of the item to retrieve.
            default (Any): The default value to return if the item is not found.

        Returns:
            T: The retrieved item or the default value.
        """

    @abstractmethod
    def __getitem__(self, item: LionIDable) -> T:
        """
        Return an item using a LionIDable identifier.

        Raises:
            KeyError: If the item ID is not found.

        Args:
            item (LionIDable): The identifier of the item to retrieve.

        Returns:
            T: The retrieved item.
        """

    @abstractmethod
    def __setitem__(self, item: LionIDable, value: T) -> None:
        """
        Add or update an item in the record.

        The value must be a Component instance.

        Args:
            item (LionIDable): The identifier of the item to add or update.
            value (T): The Component instance to add or update.
        """

    @abstractmethod
    def __contains__(self, item: LionIDable) -> bool:
        """
        Check if an item is in the record, using either an ID or an object.

        Args:
            item (LionIDable): The identifier or object to check.

        Returns:
            bool: True if the item is in the record, False otherwise.
        """

    @abstractmethod
    def __len__(self) -> int:
        """
        Return the number of items in the record.

        Returns:
            int: The number of items in the record.
        """

    @abstractmethod
    def __iter__(self) -> Iterator[T]:
        """
        Iterate over items in the record.

        Yields:
            T: The items in the record.
        """


class Ordering(ABC):
    """Represents sequencing of certain order."""

    @abstractmethod
    def __len__(self) -> int:
        """
        Return the number of item ids in the ordering.

        Or the number of orderings in another ordering.
        """

    @abstractmethod
    def __contains__(self, item: Any) -> bool:
        """Check if an item id is in the ordering."""


class Condition(ABC):
    """Represents a condition in a given context."""

    @abstractmethod
    async def applies(self, value: Any, /, *args: Any, **kwargs: Any) -> Any:
        """Asynchronously determine if the condition applies to the given value.

        Args:
            value (Any): The value to check against the condition.
            *args: Additional positional arguments.
            **kwargs: Additional keyword arguments.

        Returns:
            Any: The result of applying the condition to the value.
        """


class Actionable(ABC):
    """Represents an action that can be invoked with arguments."""

    @abstractmethod
    async def invoke(self, /, *args: Any, **kwargs: Any) -> Any:
        """
        Invoke the action asynchronously with the given arguments.

        Args:
            *args: Positional arguments for invoking the action.
            **kwargs: Keyword arguments for invoking the action.

        Returns:
            Any: The result of invoking the action.
        """


class Progressable(ABC):
    """Represents a process that can progress forward asynchronously."""

    @abstractmethod
    async def forward(self, /, *args: Any, **kwargs: Any) -> None:
        """Move the process forward asynchronously.

        Args:
            *args: Positional arguments for moving the process forward.
            **kwargs: Keyword arguments for moving the process forward.
        """


class Relatable(ABC):
    """Defines a relationship that can be established with arguments."""

    @abstractmethod
    def relate(self, /, *args: Any, **kwargs: Any) -> None:
        """Establish a relationship based on the provided arguments.

        Args:
            *args: Positional arguments for establishing the relationship.
            **kwargs: Keyword arguments for establishing the relationship.
        """


class Sendable(BaseModel, ABC):
    """Represents an object that can be sent with a sender and recipient."""

    sender: str = Field(
        "N/A",
        title="Sender",
        description=("The id of the sender node, or 'system', 'user', or 'assistant'."),
    )

    recipient: str = Field(
        "N/A",
        title="Recipient",
        description=(
            "The id of the recipient node, or 'system', 'user', or 'assistant'."
        ),
    )

    @field_validator("sender", "recipient", mode="before")
    def _validate_sender_recipient(cls, value):
        """Validate the sender and recipient fields.

        Args:
            value (Any): The value to validate.

        Returns:
            str: The validated value.

        Raises:
            LionTypeError: If the value is invalid.
        """
        if value is None:
            return "N/A"

        if value in ["system", "user", "N/A", "assistant"]:
            return value

        a = get_lion_id(value)
        if not isinstance(a, str) or len(a) != 32:
            raise LionTypeError(
                "Invalid sender or recipient value. "
                "Expected a valid node id or one of "
                "'system' or 'user'."
            )
        return a


class Executable(ABC):
    """Represents an object that can be executed with arguments."""

    @abstractmethod
    async def execute(self, /, *args: Any, **kwargs: Any) -> Any:
        """Execute the object with the given arguments asynchronously.

        Args:
            *args: Positional arguments for executing the object.
            **kwargs: Keyword arguments for executing the object.

        Returns:
            Any: The result of executing the object.
        """


class Directive(ABC):
    """Represents a directive that can be directed with arguments."""

    # @abstractmethod
    # async def direct(self, *args, **kwargs):
    #     """Direct the directive with the given arguments asynchronously.

    #     Args:
    #         *args: Positional arguments for directing the directive.
    #         **kwargs: Keyword arguments for directing the directive.
    #     """

    @property
    def class_name(self) -> str:
        """Get the class name of the directive.

        Returns:
            str: The class name of the directive.
        """
        return self._class_name()

    @classmethod
    def _class_name(cls) -> str:
        """Get the class name of the directive.

        Returns:
            str: The class name of the directive.
        """
        return cls.__name__



"""
Copyright 2024 HaiyangLi

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import inspect
from pydantic import Field
from lionagi.libs import convert, ParseUtil
from .message import RoledMessage, MessageRole


class ActionRequest(RoledMessage):
    """
    Represents a request for an action with function and arguments.

    Inherits from `RoledMessage` and provides attributes specific to action requests.

    Attributes:
        function (str): The name of the function to be called.
        arguments (dict): The keyword arguments to be passed to the function.
        action_response (str): The ID of the action response that this request corresponds to.
    """

    function: str | None = Field(
        None, description="The name of the function to be called"
    )

    arguments: dict | None = Field(
        None, description="The keyword arguments to be passed to the function"
    )

    action_response: str | None = Field(
        None,
        description="The id of the action response that this request corresponds to",
    )

    def __init__(
        self,
        function=None,
        arguments=None,
        sender=None,  # sender is the assistant who made the request
        recipient=None,  # recipient is the actionable component
        **kwargs,
    ):
        """
        Initializes the ActionRequest.

        Args:
            function (str or function, optional): The function to be called.
            arguments (dict, optional): The keyword arguments for the function.
            sender (str, optional): The sender of the request.
            recipient (str, optional): The recipient of the request.
        """
        function = function.__name__ if inspect.isfunction(function) else function
        arguments = _prepare_arguments(arguments)

        super().__init__(
            role=MessageRole.ASSISTANT,
            sender=sender,
            recipient=recipient,
            content={"action_request": {"function": function, "arguments": arguments}},
            **kwargs,
        )
        self.function = function
        self.arguments = arguments

    def is_responded(self):
        """
        Checks if the action request has been responded to.

        Returns:
            bool: True if the action request has a response, otherwise False.
        """
        return self.action_response is not None

    def clone(self, **kwargs):
        """
        Creates a copy of the current ActionRequest object with optional additional arguments.

        This method clones the current object, preserving its function and arguments.
        It also retains the original `action_response` and metadata, while allowing
        for the addition of new attributes through keyword arguments.

        Args:
            **kwargs: Optional keyword arguments to be included in the cloned object.

        Returns:
            ActionRequest: A new instance of the object with the same function, arguments,
            and additional keyword arguments.
        """
        import json

        arguments = json.dumps(self.arguments)
        request_copy = ActionRequest(
            function=self.function, arguments=json.loads(arguments), **kwargs
        )
        request_copy.action_response = self.action_response
        request_copy.metadata["origin_ln_id"] = self.ln_id
        return request_copy


def _prepare_arguments(arguments):
    """
    Prepares the arguments for the action request.

    Args:
        arguments (Any): The arguments to be prepared.

    Returns:
        dict: The prepared arguments.

    Raises:
        ValueError: If the arguments are invalid.
    """
    if not isinstance(arguments, dict):
        try:
            arguments = ParseUtil.fuzzy_parse_json(convert.to_str(arguments))
        except Exception as e:
            raise ValueError(f"Invalid arguments: {e}") from e
    if isinstance(arguments, dict):
        return arguments
    raise ValueError(f"Invalid arguments: {arguments}")



"""
Copyright 2024 HaiyangLi

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import re
import json
import contextlib

from lionagi.libs import ParseUtil
from lionagi.libs.ln_convert import strip_lower, to_dict
from lionagi.libs.ln_nested import nget

from .message import RoledMessage
from .system import System
from .instruction import Instruction
from .assistant_response import AssistantResponse
from .action_request import ActionRequest
from .action_response import ActionResponse


def create_message(
    *,
    system=None,  # system node - JSON serializable
    instruction=None,  # Instruction node - JSON serializable
    context=None,  # JSON serializable
    assistant_response=None,  # JSON
    function=None,
    arguments=None,
    func_outputs=None,
    action_request=None,  # ActionRequest node
    action_response=None,  # ActionResponse node
    images=None,  # base64 encoded image
    sender=None,  # str
    recipient=None,  # str
    requested_fields=None,  # dict[str, str]
    **kwargs,  # additional context fields
):
    # order of handling
    # action response - action request - other regular messages
    # if the message is output from function calling we will ignore other message types
    """
    Creates a message based on the provided parameters.

    Args:
        system (dict, optional): The system node (JSON serializable).
        instruction (dict, optional): The instruction node (JSON serializable).
        context (dict, optional): Additional context (JSON serializable).
        assistant_response (dict, optional): The assistant response node (JSON serializable).
        function (str, optional): The function name for action requests.
        arguments (dict, optional): The arguments for the function.
        func_outputs (Any, optional): The outputs from the function.
        action_request (ActionRequest, optional): The action request node.
        action_response (ActionResponse, optional): The action response node.
        sender (str, optional): The sender of the message.
        recipient (str, optional): The recipient of the message.
        requested_fields (dict[str, str], optional): The requested fields for the instruction.
        **kwargs: Additional context fields.

    Returns:
        RoledMessage: The constructed message based on the provided parameters.

    Raises:
        ValueError: If the parameters are invalid or missing required values.
    """

    if func_outputs or action_response:
        if not action_request:
            raise ValueError(
                "Error: please provide an corresponding action request for an action response."
            )

        if isinstance(action_response, ActionResponse):
            action_response.update_request(action_request)
            return action_response

        return ActionResponse(
            action_request=action_request,
            sender=sender,
            func_outputs=func_outputs,
        )

    if action_request:
        if not isinstance(action_request, ActionRequest):
            raise ValueError(
                "Error: action request must be an instance of ActionRequest."
            )
        return action_request

    if function:
        if not arguments:
            raise ValueError("Error: please provide arguments for the function.")
        return ActionRequest(
            function=function,
            arguments=arguments,
            sender=sender,
            recipient=recipient,
        )

    a = {
        "system": system,
        "instruction": instruction,
        "assistant_response": assistant_response,
    }

    a = {k: v for k, v in a.items() if v is not None}

    if not len(a) == 1:
        raise ValueError("Error: Message can only have one role")

    if not func_outputs:
        for k, v in a.items():
            if isinstance(v, RoledMessage):
                if isinstance(v, Instruction):
                    if context:
                        v._add_context(context)
                    if requested_fields:
                        v._update_requested_fields(requested_fields)
                return v

    if system:
        return System(system=system, sender=sender, recipient=recipient)

    elif assistant_response:
        return AssistantResponse(
            assistant_response=assistant_response,
            sender=sender,
            recipient=recipient,
        )

    else:
        if images:
            images = images if isinstance(images, list) else [images]

        return Instruction(
            instruction=instruction,
            context=context,
            sender=sender,
            recipient=recipient,
            requested_fields=requested_fields,
            images=images,
            **kwargs,
        )


def _parse_action_request(response):
    """
    Parses an action request from the response.

    Args:
        response (dict): The response containing the action request.

    Returns:
        list[ActionRequest] or None: A list of action requests or None if invalid.

    Raises:
        ActionError: If the action request is invalid.
    """
    message = to_dict(response) if not isinstance(response, dict) else response
    content_ = None

    if strip_lower(nget(message, ["content"])) == "none":
        content_ = _handle_action_request(message)

    elif nget(message, ["content", "tool_uses"], None):
        content_ = message["content"]["tool_uses"]

    else:
        json_block_pattern = re.compile(r"```json\n({.*?tool_uses.*?})\n```", re.DOTALL)

        # Find the JSON block in the text
        match = json_block_pattern.search(str(message["content"]))
        if match:
            json_block = match.group(1)
            parsed_json = json.loads(json_block)
            if "tool_uses" in parsed_json:
                content_ = parsed_json["tool_uses"]
            elif "actions" in parsed_json:
                content_ = parsed_json["actions"]
            else:
                content_ = []

    if isinstance(content_, dict):
        content_ = [content_]

    if isinstance(content_, list) and not content_ == []:
        outs = []
        for func_calling in content_:
            if "recipient_name" in func_calling:
                func_calling["action"] = func_calling["recipient_name"].split(".")[1]
                func_calling["arguments"] = func_calling["parameters"]
            elif "function" in func_calling:
                func_calling["action"] = func_calling["function"]
                if "parameters" in func_calling:
                    func_calling["arguments"] = func_calling["parameters"]
                elif "arguments" in func_calling:
                    func_calling["arguments"] = func_calling["arguments"]

            msg = ActionRequest(
                function=func_calling["action"]
                .replace("action_", "")
                .replace("recipient_", ""),
                arguments=func_calling["arguments"],
            )
            outs.append(msg)
        return outs

    else:
        try:
            _content = to_dict(message["content"])
            if "action_request" in _content:
                content_ = _content["action_request"]

            if isinstance(content_, dict):
                content_ = [content_]

            if isinstance(content_, list):
                outs = []
                for func_calling in content_:
                    if "function" in func_calling:
                        func_calling["action"] = func_calling["function"]
                        if "parameters" in func_calling:
                            func_calling["arguments"] = func_calling["parameters"]
                        elif "arguments" in func_calling:
                            func_calling["arguments"] = func_calling["arguments"]
                    msg = ActionRequest(
                        function=func_calling["action"]
                        .replace("action_", "")
                        .replace("recipient_", ""),
                        arguments=func_calling["arguments"],
                    )
                    outs.append(msg)
                return outs
        except:
            return None
    return None


def _handle_action_request(response):
    """
    Handles the action request parsing from the response.

    Args:
        response (dict): The response containing the action request details.

    Returns:
        list[dict]: A list of function call details.

    Raises:
        ValueError: If the response message is invalid.
    """
    try:
        tool_count = 0
        func_list = []
        while tool_count < len(response["tool_calls"]):
            _path = ["tool_calls", tool_count, "type"]

            if nget(response, _path) == "function":
                _path1 = ["tool_calls", tool_count, "function", "name"]
                _path2 = ["tool_calls", tool_count, "function", "arguments"]

                func_content = {
                    "action": f"action_{nget(response, _path1)}",
                    "arguments": nget(response, _path2),
                }
                func_list.append(func_content)
            tool_count += 1
        return func_list
    except:
        raise ValueError(
            "Response message must be one of regular response or function calling"
        )



from lionagi.libs.ln_func_call import lcall
from lionagi.core.rule.base import Rule


"""
rule config schema 

{
    rule_name: {
        "fields: [],
        "config": {}, 
        ...
    }
}
"""


class RuleBook:

    def __init__(
        self,
        rules: dict[str, Rule] | list[Rule] = None,
        ruleorder: list[str] = None,
        rule_config: dict[str, dict] = None,
    ):
        self.rules = rules
        self.ruleorder = ruleorder
        self.rule_config = rule_config or {k: {} for k in self.ruleorder}

    @property
    def _all_applied_log(self):
        """return all applied logs from all rules in the rulebook"""
        return lcall(self.rules.values(), lambda x: x.applied_log, flatten=True)

    @property
    def _all_invoked_log(self):
        """return all invoked logs from all rules in the rulebook"""
        return lcall(self.rules.values(), lambda x: x.invoked_log, flatten=True)

    def __getitem__(self, key: str) -> Rule:
        return self.rules[key]

    # def add_rule(self, rule_name: str, rule: Rule, config: dict = None):
    #     if rule_name in self.rules:
    #         raise ValueError(f"Rule '{rule_name}' already exists.")
    #     self.rules[rule_name] = rule
    #     self.ruleorder.append(rule_name)
    #     self.rule_config[rule_name] = config or {}

    # def remove_rule(self, rule_name: str):
    #     if rule_name not in self.rules:
    #         raise ValueError(f"Rule '{rule_name}' does not exist.")
    #     del self.rules[rule_name]
    #     self.ruleorder.remove(rule_name)
    #     del self.rule_config[rule_name]

    # def update_rule_config(self, rule_name: str, config: dict):
    #     if rule_name not in self.rules:
    #         raise ValueError(f"Rule '{rule_name}' does not exist.")
    #     self.rule_config[rule_name] = config

    # def list_rules(self) -> list[str]:
    #     return self.ruleorder

    # def get_rule_details(self, rule_name: str) -> dict:
    #     if rule_name not in self.rules:
    #         raise ValueError(f"Rule '{rule_name}' does not exist.")
    #     return {
    #         "rule": self.rules[rule_name],
    #         "config": self.rule_config[rule_name]
    #     }

    # async def validate_data(self, data: Any) -> bool:
    #     for rule in self.rules.values():
    #         if not await rule.validate(data):
    #             return False
    #     return True

    # def export_rulebook(self, filepath: str):
    #     import json
    #     with open(filepath, 'w') as f:
    #         json.dump({
    #             "rules": list(self.rules.keys()),
    #             "ruleorder": self.ruleorder,
    #             "rule_config": self.rule_config
    #         }, f)

    # @classmethod
    # def import_rulebook(cls, filepath: str) -> 'RuleBook':
    #     import json
    #     with open(filepath, 'r') as f:
    #         config = json.load(f)
    #     rules = {name: Rule() for name in config["rules"]}
    #     return cls(rules=rules, ruleorder=config["ruleorder"], rule_config=config["rule_config"])

    # def enable_rule(self, rule_name: str, enable: bool = True):
    #     if rule_name not in self.rules:
    #         raise ValueError(f"Rule '{rule_name}' does not exist.")
    #     self.rules[rule_name].enabled = enable

    # def log_rule_application(self, rule_name: str, data: Any):
    #     if rule_name not in self.rules:
    #         raise ValueError(f"Rule '{rule_name}' does not exist.")
    #     log_entry = {
    #         "rule": rule_name,
    #         "data": data,
    #         "timestamp": SysUtil.get_timestamp()
    #     }
    #     # Append log_entry to a log file or a logging system



#TODO
# import unittest
# from unittest.mock import MagicMock, patch
# from lionagi.core.graph.graph import Graph
# from lionagi.core.generic.node import Node
# from lionagi.core.generic.edge import Edge
#
# class TestGraph(unittest.TestCase):
#
#     def setUp(self):
#         self.graph = Graph()
#         self.node1 = Node(id_="node1", content="Node 1 content")
#         self.node2 = Node(id_="node2", content="Node 2 content")
#         self.node3 = Node(id_="node3", content="Node 3 content")
#         self.graph.add_node(self.node1)
#         self.graph.add_node(self.node2)
#         self.graph.add_node(self.node3)
#
#     def test_graph_heads(self):
#         self.graph.relate_nodes(self.node1, self.node2)
#         self.assertEqual(["node1"], self.graph.graph_heads)
#
#
#     def test_acyclic(self):
#         self.graph.relate_nodes(self.node1, self.node2)
#         self.assertTrue(self.graph.acyclic)
#
#         self.graph.relate_nodes(self.node2, self.node1)  # Creating a cycle
#         self.assertFalse(self.graph.acyclic)
#
#     @patch("lionagi.libs.SysUtil.check_import")
#     def test_to_networkx_success(self, mock_check_import):
#         mock_check_import.return_value = None
#         with patch("networkx.DiGraph") as mock_digraph:
#             mock_graph = MagicMock()
#             mock_digraph.return_value = mock_graph
#             result = self.graph.to_networkx()
#             self.assertEqual(result, mock_graph)
#
#     @patch("lionagi.libs.SysUtil.check_import")
#     def test_to_networkx_empty_graph(self, mock_check_import):
#         mock_check_import.return_value = None
#         with patch("networkx.DiGraph") as mock_digraph:
#             mock_graph = MagicMock()
#             mock_digraph.return_value = mock_graph
#
#             self.graph.internal_nodes = {}
#             result = self.graph.to_networkx()
#
#             self.assertEqual(result, mock_graph)
#             mock_check_import.assert_called_once_with("networkx")
#             mock_digraph.assert_called_once()
#             mock_graph.add_node.assert_not_called()
#             mock_graph.add_edge.assert_not_called()
#
#     def test_add_node(self):
#         new_node = Node(id_="node4", content="Node 4 content")
#         self.graph.add_node(new_node)
#         self.assertIn("node4", self.graph.internal_nodes)
#
#     def test_remove_node(self):
#         self.graph.remove_node(self.node1)
#         self.assertNotIn("node1", self.graph.internal_nodes)
#
#     def test_clear(self):
#         self.graph.clear()
#         self.assertEqual(len(self.graph.internal_nodes), 0)
#         self.assertTrue(self.graph.is_empty)
#
# if __name__ == "__main__":
#     unittest.main()



"""
Copyright 2024 HaiyangLi

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from .base import BaseUnitForm, Field


class PredictTemplate(BaseUnitForm):
    """
    A template for predicting the next sentence(s) based on given instructions and context.

    Inherits from `BaseUnitForm` and adds fields and methods specific to prediction tasks.

    Attributes:
        confidence_score (float | None): A numeric score between 0 to 1 formatted to 2 decimal places, indicating confidence in the prediction.
        reason (str | None): A brief reason for the given output.
        template_name (str): The name of the template.
        num_sentences (int): The number of sentences to predict.
        prediction (None | str | list): The predicted sentence(s) or desired output.
        assignment (str): A string representing the task assignment for the prediction.
    """

    confidence_score: float | None = Field(
        None,
        description="a numeric score between 0 to 1 formatted in num:0.2f, 1 being very confident and 0 being not confident at all, just guessing",
        validation_kwargs={
            "upper_bound": 1,
            "lower_bound": 0,
            "num_type": float,
            "precision": 2,
        },
    )

    reason: str | None = Field(
        None,
        description="brief reason for the given output, format: This is my best response because ...",
    )

    template_name: str = "predict_template"

    num_sentences: int = Field(2, description="the number of sentences to predict")

    prediction: None | str | list = Field(
        None,
        description="the predicted sentence(s) or desired output",
    )

    assignment: str = "task -> prediction"

    @property
    def answer(self):
        """
        Gets the prediction attribute.

        Returns:
            None | str | list: The predicted sentence(s) or desired output.
        """
        return getattr(self, "prediction", None)

    def __init__(
        self,
        *,
        instruction=None,
        context=None,
        num_sentences=2,
        confidence_score=False,
        reason=False,
        **kwargs,
    ):
        """
        Initializes a new instance of the PredictTemplate class.

        Args:
            instruction (str, optional): Additional instructions for the prediction.
            context (str, optional): Additional context for the prediction.
            num_sentences (int, optional): The number of sentences to predict. Defaults to 2.
            confidence_score (bool, optional): Whether to include a confidence score. Defaults to False.
            reason (bool, optional): Whether to include a reasoning field. Defaults to False.
            **kwargs: Additional keyword arguments.
        """

        super().__init__(**kwargs)

        self.num_sentences = num_sentences

        self.task = f"""
predict the next sentence(s) according to the following constraints
1. number of sentences: {self.num_sentences}
2. additional objective , {instruction or "N/A"}
3. additional information, {context or "N/A"}
"""

        if reason:
            self.append_to_request("reason")

        if confidence_score:
            self.append_to_request("confidence_score")



"""
Copyright 2024 HaiyangLi

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

"""
This module contains the ToolManager class, which manages tools in the system.
It allows registering, invoking, and retrieving schemas of tools. Tools can be
registered individually or in batches, and invoked using function names, JSON
strings, or specialized objects.
"""

import inspect
from functools import singledispatchmethod
from typing import Any, Callable, List, Union, Tuple
from lionagi.libs import ParseUtil
from lionagi.libs.ln_convert import to_list, to_dict
from lionagi.libs.ln_func_call import lcall
from lionagi.core.collections.abc import Actionable
from lionagi.core.action.function_calling import FunctionCalling
from lionagi.core.action.tool import Tool, TOOL_TYPE


class ToolManager(Actionable):
    """
    Manages tools in the system. Provides functionality to register tools,
    invoke them based on various input formats, and retrieve tool schemas.
    """

    def __init__(self, registry: dict[str, Tool] = None) -> None:
        """
        Initializes a new instance of ToolManager.

        Args:
            registry (dict[str, Tool], optional): A dictionary to store registered tools.
                Defaults to an empty dictionary.
        """
        self.registry = registry or {}

    def __contains__(self, tool) -> bool:
        if isinstance(tool, Tool):
            return tool.name in self.registry

        elif isinstance(tool, str):
            return tool in self.registry

        elif inspect.isfunction(tool):
            return tool.__name__ in self.registry

        return False

    def _register_tool(self, tool: Tool | Callable, update: bool = False) -> bool:
        """
        Registers a single tool or multiple tools based on the input type.

        Args:
            tool (Any): Can be a single Tool, a list of Tools, a callable, or other forms.

        Raises:
            TypeError: If the tools argument type is unsupported.
        """
        if not update and tool in self:
            raise ValueError(f"Function {tool.name} is already registered.")
        if isinstance(tool, Callable):
            tool = func_to_tool(tool)
            tool = tool[0] if isinstance(tool, list) and len(tool) == 1 else tool
        if not isinstance(tool, Tool):
            raise TypeError("Please register a Tool object.")
        self.registry[tool.name] = tool
        return True

    def update_tools(self, tools: list | Tool | Callable):
        if isinstance(tools, Tool) or isinstance(tools, Callable):
            return self._register_tool(tools, update=True)
        else:
            return all(lcall(tools, self._register_tool, update=True))

    def register_tools(self, tools: list | Tool | Callable):
        """
        Registers multiple Tools.

        Args:
            tools (list): The list of Tools to register.

        Returns:
            bool: True if all tools are successfully registered.
        """
        if isinstance(tools, Tool) or isinstance(tools, Callable):
            return self._register_tool(tools)
        else:
            return all(lcall(tools, self._register_tool))

    async def invoke(self, func_calling=None):
        """
        Invokes a function based on the provided function calling description.

        Args:
            func_calling (Any, optional): The function calling description, which can be:
                - tuple: (name, kwargs)
                - dict: {"function": name, "arguments": kwargs}
                - str: JSON string of dict format
                - ActionRequest object
                - FunctionCalling object

        Returns:
            Any: The result of the function call.

        Raises:
            ValueError: If func_calling is None or the function is not registered.
        """
        if not func_calling:
            raise ValueError("func_calling is required.")

        if not isinstance(func_calling, FunctionCalling):
            func_calling = FunctionCalling.create(func_calling)

        if func_calling.func_name in self.registry:
            return await self.registry[func_calling.func_name].invoke(
                func_calling=func_calling
            )

        raise ValueError(f"Function {func_calling.func_name} is not registered.")

    @property
    def _schema_list(self) -> list[dict[str, Any]]:
        """
        Lists all tool schemas currently registered in the ToolManager.

        Returns:
            list[dict[str, Any]]: A list of tool schemas.
        """
        return [tool.schema_ for tool in self.registry.values()]

    def get_tool_schema(self, tools: TOOL_TYPE, **kwargs) -> dict:
        """
        Retrieves the schema for a specific tool or all tools based on the input.

        Args:
            tools (TOOL_TYPE): Can be a boolean, specific tool name, Tool, or list of tools.
            **kwargs: Additional keyword arguments to be merged with tool schema.

        Returns:
            dict: Combined tool schema and kwargs.
        """
        if isinstance(tools, bool):
            tool_kwarg = {"tools": self._schema_list}
            return tool_kwarg | kwargs

        else:
            tool_kwarg = {"tools": self._get_tool_schema(tools)}
            return tool_kwarg | kwargs

    @singledispatchmethod
    def _get_tool_schema(self, tool: Any) -> dict:
        """
        Retrieves the schema for a specific tool based on its type.

        Args:
            tool (Any): The tool descriptor, can be a dict, Tool, str, or list.

        Raises:
            TypeError: If the tool type is unsupported.
        """
        raise TypeError(f"Unsupported type {type(tool)}")

    @_get_tool_schema.register(dict)
    def _(self, tool):
        """
        Assumes that the tool is a schema and returns it.

        Args:
            tool (dict): The tool schema.

        Returns:
            dict: The tool schema.
        """
        return tool

    @_get_tool_schema.register(Tool)
    def _(self, tool):
        """
        Retrieves the schema for a Tool object.

        Args:
            tool (Tool): The Tool object.

        Returns:
            dict: The tool schema.

        Raises:
            ValueError: If the Tool is not registered.
        """
        if tool.name in self.registry:
            return self.registry[tool.name].schema_
        else:
            err_msg = f"Function {tool.name} is not registered."
            raise ValueError(err_msg)

    @_get_tool_schema.register(str)
    def _(self, tool):
        """
        Assumes that the tool is a name and retrieves its schema.

        Args:
            tool (str): The tool name.

        Returns:
            dict: The tool schema.

        Raises:
            ValueError: If the tool name is not registered.
        """
        if tool in self.registry:
            return self.registry[tool].schema_
        else:
            err_msg = f"Function {tool} is not registered."
            raise ValueError(err_msg)

    @_get_tool_schema.register(list)
    def _(self, tools):
        """
        Retrieves the schema for a list of tools.

        Args:
            tools (list): The list of tools.

        Returns:
            list: A list of tool schemas.
        """
        return lcall(tools, self._get_tool_schema)

    def parse_tool(self, tools: TOOL_TYPE, **kwargs) -> dict:
        """
        Parses and merges tool schemas based on the provided tool descriptors.

        Args:
            tools (TOOL_TYPE): The tools to parse, can be a single tool or list.
            **kwargs: Additional keyword arguments to be merged with the tool schema.

        Returns:
            dict: The merged tool schema and additional arguments.
        """
        if tools:
            if isinstance(tools, bool):
                tool_kwarg = {"tools": self._schema_list}
                kwargs = tool_kwarg | kwargs
            else:
                tools = to_list(tools) if not isinstance(tools, list) else [tools]
                tool_kwarg = {"tools": lcall(tools, self._get_tool_schema)}
                kwargs = tool_kwarg | kwargs

        return kwargs

    @staticmethod
    def parse_tool_request(response: dict) -> Tuple[str, dict]:
        """
        Parses a tool request from a given response dictionary.

        Args:
            response (dict): The response data containing the tool request.

        Returns:
            Tuple[str, dict]: The function name and its arguments.

        Raises:
            ValueError: If the response is not a valid function call.
        """
        try:
            func = response["action"][7:]
            args = to_dict(response["arguments"])
            return func, args
        except Exception:
            try:
                func = response["recipient_name"].split(".")[-1]
                args = response["parameters"]
                return func, args
            except:
                raise ValueError("response is not a valid function call")


def func_to_tool(
    func_: Union[Callable, List[Callable]],
    parser: Union[Callable, List[Callable]] = None,
    docstring_style: str = "google",
    **kwargs,
) -> List[Tool]:
    """
    Converts functions to Tool objects, optionally associating parsers with each function
    and applying a specified docstring parsing style to generate tool schemas.

    Args:
        func_ (Callable | list[Callable]): The function(s) to convert into tool(s).
        parser (Callable | list[Callable], optional): Parser(s) to associate with
            the function(s). If a list is provided, it should match the length of func_.
        docstring_style (str, optional): The style of the docstring parser to use when
            generating tool schemas. Defaults to "google".

    Returns:
        list[Tool]: A list of Tool objects created from the provided function(s).

    Raises:
        ValueError: If the length of the parsers does not match the length of the
            functions when both are provided as lists.

    Examples:
        # Convert a single function with a custom parser
        tools = func_to_tool(my_function, my_parser)

        # Convert multiple functions without parsers
        tools = func_to_tool([func_one, func_two])

        # Convert multiple functions with multiple parsers
        tools = func_to_tool([func_one, func_two], [parser_one, parser_two])
    """
    fs = []
    funcs = to_list(func_, flatten=True, dropna=True)
    parsers = to_list(parser, flatten=True, dropna=True)

    if parser:
        if len(funcs) != len(parsers) != 1:
            raise ValueError(
                "Length of parser must match length of func. Except if you only pass one"
            )

        for idx in range(len(funcs)):
            f_ = lambda _f: Tool(
                function=_f,
                schema_=ParseUtil._func_to_schema(_f, style=docstring_style),
                parser=parsers[idx] if len(parsers) > 1 else parsers[0],
                **kwargs,
            )

            fs.append(f_)

    else:
        fs = lcall(
            funcs,
            lambda _f: Tool(
                function=_f,
                schema_=ParseUtil._func_to_schema(_f, style=docstring_style),
                **kwargs,
            ),
        )

    return fs



#TODO

# import unittest
# from unittest.mock import MagicMock, patch
# from lionagi.core.tool.structure import *
#
#
# class TestBaseStructure(unittest.TestCase):
#     def setUp(self):
#         self.structure = BaseStructure()
#         self.node1 = BaseNode(id_="node1")
#         self.node2 = BaseNode(id_="node2")
#         self.node3 = BaseNode(id_="node3")
#         self.edge1 = Edge(id_="edge1", source_node_id="node1", target_node_id="node2")
#         self.edge2 = Edge(id_="edge2", source_node_id="node2", target_node_id="node3")
#
#     def test_node_edges_property(self):
#         self.node1.in_relations = {"edge1": self.edge1}
#         self.node1.out_relations = {"edge2": self.edge2}
#         self.structure.structure_nodes = {"node1": self.node1}
#         expected_result = {
#             "node1": {"in": {"edge1": self.edge1}, "out": {"edge2": self.edge2}}
#         }
#         self.assertEqual(self.structure.node_edges, expected_result)
#
#     def test_get_node_edges_with_node(self):
#         self.node1.out_relations = {"edge1": self.edge1}
#         self.node1.in_relations = {"edge2": self.edge2}
#         self.structure.structure_nodes = {"node1": self.node1}
#         self.assertEqual(
#             self.structure.get_node_edges(self.node1, direction="out"), [self.edge1]
#         )
#         self.assertEqual(
#             self.structure.get_node_edges(self.node1, direction="in"), [self.edge2]
#         )
#         self.assertEqual(
#             self.structure.get_node_edges(self.node1, direction="all"),
#             [self.edge2, self.edge1],
#         )
#
#     def test_get_node_edges_without_node(self):
#         self.structure.structure_edges = {"edge1": self.edge1, "edge2": self.edge2}
#         self.assertEqual(self.structure.get_node_edges(), [self.edge1, self.edge2])
#
#     def test_get_node_edges_node_not_found(self):
#         invalid_node = BaseNode(id_="invalid_node")
#         try:
#             self.structure.get_node_edges(invalid_node)
#         except KeyError as e:
#             self.assertEqual(str(e), f"node {invalid_node.id_} is not found")
#
#     def test_has_structure_edge_with_edge_object(self):
#         self.structure.structure_edges = {"edge1": self.edge1}
#         self.assertTrue(self.structure.has_structure_edge(self.edge1))
#
#     def test_has_structure_edge_with_edge_id(self):
#         self.structure.structure_edges = {"edge1": self.edge1}
#         self.assertTrue(self.structure.has_structure_edge("edge1"))
#
#     def test_get_structure_edge_with_edge_object(self):
#         self.structure.structure_edges = {"edge1": self.edge1}
#         self.assertEqual(self.structure.get_structure_edge(self.edge1), self.edge1)
#
#     def test_get_structure_edge_with_edge_id(self):
#         self.structure.structure_edges = {"edge1": self.edge1}
#         self.assertEqual(self.structure.get_structure_edge("edge1"), self.edge1)
#
#     def test_add_structure_edge_success(self):
#         self.structure.structure_nodes = {"node1": self.node1, "node2": self.node2}
#         self.structure.add_structure_edge(self.edge1)
#         self.assertEqual(self.structure.structure_edges, {"edge1": self.edge1})
#         self.assertEqual(self.node1.out_relations, {"node2": self.edge1})
#         self.assertEqual(self.node2.in_relations, {"node1": self.edge1})
#
#     def test_add_structure_edge_failure(self):
#         with self.assertRaises(ValueError):
#             self.structure.add_structure_edge(self.edge1)
#
#     def test_remove_structure_edge_with_edge_object(self):
#         self.structure.structure_edges = {"edge1": self.edge1}
#         self.structure.structure_nodes = {"node1": self.node1, "node2": self.node2}
#         self.node1.out_relations = {"node2": self.edge1}
#         self.node2.in_relations = {"node1": self.edge1}
#         self.structure.remove_structure_edge(self.edge1)
#         self.assertEqual(self.structure.structure_edges, {})
#         self.assertEqual(self.node1.out_relations, {})
#         self.assertEqual(self.node2.in_relations, {})
#
#     def test_remove_structure_edge_with_edge_id(self):
#         self.structure.structure_edges = {"edge1": self.edge1}
#         self.structure.structure_nodes = {"node1": self.node1, "node2": self.node2}
#         self.node1.out_relations = {"node2": self.edge1}
#         self.node2.in_relations = {"node1": self.edge1}
#         self.structure.remove_structure_edge("edge1")
#         self.assertEqual(self.structure.structure_edges, {})
#         self.assertEqual(self.node1.out_relations, {})
#         self.assertEqual(self.node2.in_relations, {})
#
#     def test_remove_structure_node_success(self):
#         self.structure.structure_nodes = {"node1": self.node1, "node2": self.node2}
#         self.structure.structure_edges = {"edge1": self.edge1}
#         self.node1.out_relations = {"node2": self.edge1}
#         self.node2.in_relations = {"node1": self.edge1}
#         self.structure.remove_structure_node(self.node1)
#         self.assertEqual(self.structure.structure_nodes, {"node2": self.node2})
#         self.assertEqual(self.structure.structure_edges, {})
#
#     def test_remove_structure_node_failure(self):
#         with self.assertRaises(ValueError):
#             self.structure.remove_structure_node(self.node1)
#
#     def test_add_structure_node_with_base_node(self):
#         self.structure.add_structure_node(self.node1)
#         self.assertEqual(self.structure.structure_nodes, {"node1": self.node1})
#
#     def test_add_structure_node_with_list(self):
#         self.structure.add_structure_node([self.node1, self.node2])
#         self.assertEqual(
#             self.structure.structure_nodes, {"node1": self.node1, "node2": self.node2}
#         )
#
#     def test_add_structure_node_with_dict(self):
#         self.structure.add_structure_node({"node1": self.node1, "node2": self.node2})
#         self.assertEqual(
#             self.structure.structure_nodes, {"node1": self.node1, "node2": self.node2}
#         )
#
#     def test_add_structure_node_unsupported_type(self):
#         with self.assertRaises(NotImplementedError):
#             self.structure.add_structure_node(1)
#
#     def test_get_structure_node_with_node_id(self):
#         self.structure.structure_nodes = {"node1": self.node1}
#         self.assertEqual(self.structure.get_structure_node("node1"), self.node1)
#
#     def test_get_structure_node_with_base_node(self):
#         self.structure.structure_nodes = {"node1": self.node1}
#         self.assertEqual(self.structure.get_structure_node(self.node1), self.node1)
#
#     def test_get_structure_node_with_list(self):
#         self.structure.structure_nodes = {"node1": self.node1, "node2": self.node2}
#         self.assertEqual(
#             self.structure.get_structure_node(["node1", "node2"]),
#             [self.node1, self.node2],
#         )
#
#     def test_pop_structure_node_with_node_id(self):
#         self.structure.structure_nodes = {"node1": self.node1}
#         self.assertEqual(self.structure.pop_structure_node("node1"), self.node1)
#         self.assertEqual(self.structure.structure_nodes, {})
#
#     def test_pop_structure_node_with_base_node(self):
#         self.structure.structure_nodes = {"node1": self.node1}
#         self.assertEqual(self.structure.pop_structure_node(self.node1), self.node1)
#         self.assertEqual(self.structure.structure_nodes, {})
#
#     def test_pop_structure_node_with_list(self):
#         self.structure.structure_nodes = {"node1": self.node1, "node2": self.node2}
#         self.assertEqual(
#             self.structure.pop_structure_node(["node1", "node2"]),
#             [self.node1, self.node2],
#         )
#         self.assertEqual(self.structure.structure_nodes, {})
#
#     def test_pop_structure_node_unsupported_type(self):
#         with self.assertRaises(NotImplementedError):
#             self.structure.pop_structure_node(1)
#
#     def test_has_structure_node_with_node_id(self):
#         self.structure.structure_nodes = {"node1": self.node1}
#         self.assertTrue(self.structure.has_structure_node("node1"))
#         self.assertFalse(self.structure.has_structure_node("node2"))
#
#     def test_has_structure_node_with_base_node(self):
#         self.structure.structure_nodes = {"node1": self.node1}
#         self.assertTrue(self.structure.has_structure_node(self.node1))
#         self.assertFalse(self.structure.has_structure_node(self.node2))
#
#     def test_has_structure_node_with_list(self):
#         self.structure.structure_nodes = {"node1": self.node1, "node2": self.node2}
#         self.assertTrue(self.structure.has_structure_node(["node1", "node2"]))
#         self.assertFalse(self.structure.has_structure_node(["node1", "node3"]))
#
#     def test_is_empty_property(self):
#         self.assertTrue(self.structure.is_empty)
#         self.structure.structure_nodes = {"node1": self.node1}
#         self.assertFalse(self.structure.is_empty)
#
#     def test_clear_method(self):
#         self.structure.structure_nodes = {"node1": self.node1}
#         self.structure.structure_edges = {"edge1": self.edge1}
#         self.structure.clear()
#         self.assertEqual(self.structure.structure_nodes, {})
#         self.assertEqual(self.structure.structure_edges, {})
#
#
# if __name__ == "__main__":
#     unittest.main()



"""
Copyright 2024 HaiyangLi

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from lionagi.core.unit.unit import Unit
from lionagi.core.session.branch import Branch


async def chat(
    instruction=None,
    context=None,
    system=None,
    sender=None,
    recipient=None,
    branch=None,
    form=None,
    confidence_score=None,
    reason=False,
    **kwargs,
):
    """
    Performs a chat operation using the specified parameters.

    Args:
        instruction (str, optional): The instruction for the chat.
        context (Any, optional): The context to perform the instruction on.
        system (Any, optional): The system context for the chat.
        sender (str, optional): The sender of the instruction.
        recipient (str, optional): The recipient of the instruction.
        branch (Branch, optional): The branch to use for the chat.
        form (Any, optional): The form to create instruction from.
        confidence_score (float, optional): The confidence score for the operation.
        reason (bool, optional): Whether to include a reason for the operation.
        **kwargs: Additional keyword arguments for the chat operation.

    Returns:
        Any: The result of the chat operation.
    """
    branch = branch or Branch()
    unit = Unit(branch)

    return await unit.chat(
        instruction=instruction,
        context=context,
        system=system,
        sender=sender,
        recipient=recipient,
        form=form,
        confidence_score=confidence_score,
        reason=reason,
        branch=branch,
        **kwargs,
    )


async def select(
    instruction=None,
    context=None,
    system=None,
    sender=None,
    recipient=None,
    choices=None,
    branch=None,
    form=None,
    confidence_score=None,
    reason=False,
    **kwargs,
):
    """
    Performs a select operation using the specified parameters.

    Args:
        instruction (str, optional): The instruction for the selection.
        context (Any, optional): The context to perform the instruction on.
        system (Any, optional): The system context for the selection.
        sender (str, optional): The sender of the instruction.
        recipient (str, optional): The recipient of the instruction.
        choices (list, optional): The choices for the selection.
        branch (Branch, optional): The branch to use for the selection.
        form (Any, optional): The form to create instruction from.
        confidence_score (float, optional): The confidence score for the operation.
        reason (bool, optional): Whether to include a reason for the operation.
        **kwargs: Additional keyword arguments for the selection operation.

    Returns:
        Any: The result of the selection operation.
    """
    branch = branch or Branch()
    unit = Unit(branch)

    return await unit.select(
        instruction=instruction,
        context=context,
        system=system,
        sender=sender,
        recipient=recipient,
        choices=choices,
        form=form,
        confidence_score=confidence_score,
        reason=reason,
        **kwargs,
    )


async def predict(
    instruction=None,
    context=None,
    system=None,
    sender=None,
    recipient=None,
    branch=None,
    form=None,
    confidence_score=None,
    reason=False,
    num_sentences=1,
    **kwargs,
):
    """
    Performs a predict operation using the specified parameters.

    Args:
        instruction (str, optional): The instruction for the prediction.
        context (Any, optional): The context to perform the instruction on.
        system (Any, optional): The system context for the prediction.
        sender (str, optional): The sender of the instruction.
        recipient (str, optional): The recipient of the instruction.
        branch (Branch, optional): The branch to use for the prediction.
        form (Any, optional): The form to create instruction from.
        confidence_score (float, optional): The confidence score for the operation.
        reason (bool, optional): Whether to include a reason for the operation.
        **kwargs: Additional keyword arguments for the prediction operation.

    Returns:
        Any: The result of the prediction operation.
    """
    branch = branch or Branch()
    unit = Unit(branch)

    return await unit.predict(
        instruction=instruction,
        context=context,
        system=system,
        sender=sender,
        recipient=recipient,
        form=form,
        confidence_score=confidence_score,
        reason=reason,
        num_sentences=num_sentences,
        **kwargs,
    )


async def act(
    instruction=None,
    context=None,
    system=None,
    sender=None,
    recipient=None,
    branch=None,
    form=None,
    confidence_score=None,
    reason=False,
    **kwargs,
):
    """
    Performs an act operation using the specified parameters.

    Args:
        instruction (str, optional): The instruction for the action.
        context (Any, optional): The context to perform the instruction on.
        system (Any, optional): The system context for the action.
        sender (str, optional): The sender of the instruction.
        recipient (str, optional): The recipient of the instruction.
        branch (Branch, optional): The branch to use for the action.
        form (Any, optional): The form to create instruction from.
        confidence_score (float, optional): The confidence score for the operation.
        reason (bool, optional): Whether to include a reason for the operation.
        **kwargs: Additional keyword arguments for the act operation.

    Returns:
        Any: The result of the act operation.
    """
    branch = branch or Branch()
    unit = Unit(branch)

    return await unit.act(
        instruction=instruction,
        context=context,
        system=system,
        sender=sender,
        recipient=recipient,
        form=form,
        confidence_score=confidence_score,
        reason=reason,
        **kwargs,
    )


async def score(
    instruction=None,
    context=None,
    system=None,
    sender=None,
    recipient=None,
    branch=None,
    form=None,
    confidence_score=None,
    reason=False,
    score_range=None,
    include_endpoints=None,
    num_digit=None,
    **kwargs,
):
    """
    Asynchronously performs a scoring task within a given context.

    Args:
        instruction (str, optional): Additional instruction for the scoring task.
        context (Any, optional): Context to perform the scoring task on.
        system (str, optional): System message to use for the scoring task.
        sender (str, optional): Sender of the instruction. Defaults to None.
        recipient (str, optional): Recipient of the instruction. Defaults to None.
        branch (Branch, optional): Branch to perform the task within. Defaults to a new Branch.
        form (Form, optional): Form to create the instruction from. Defaults to None.
        confidence_score (bool, optional): Flag to include a confidence score. Defaults to None.
        reason (bool, optional): Flag to include a reason for the scoring. Defaults to False.
        score_range (tuple, optional): Range for the score. Defaults to None.
        include_endpoints (bool, optional): Flag to include endpoints in the score range. Defaults to None.
        num_digit (int, optional): Number of decimal places for the score. Defaults to None.
        **kwargs: Additional keyword arguments for further customization.

    Returns:
        Any: The result of the scoring task.
    """

    branch = branch or Branch()
    unit = Unit(branch)

    return await unit.score(
        instruction=instruction,
        context=context,
        system=system,
        sender=sender,
        recipient=recipient,
        form=form,
        confidence_score=confidence_score,
        reason=reason,
        score_range=score_range,
        include_endpoints=include_endpoints,
        num_digit=num_digit,
        **kwargs,
    )


async def plan(
    instruction=None,
    context=None,
    system=None,
    sender=None,
    recipient=None,
    branch=None,
    form=None,
    confidence_score=None,
    reason=False,
    num_step=3,
    **kwargs,
):
    """
    Asynchronously generates a step-by-step plan within a given context.

    Args:
        instruction (str, optional): Additional instruction for the planning task.
        context (Any, optional): Context to perform the planning task on.
        system (str, optional): System message to use for the planning task.
        sender (str, optional): Sender of the instruction. Defaults to None.
        recipient (str, optional): Recipient of the instruction. Defaults to None.
        branch (Branch, optional): Branch to perform the task within. Defaults to a new Branch.
        form (Form, optional): Form to create the instruction from. Defaults to None.
        confidence_score (bool, optional): Flag to include a confidence score. Defaults to None.
        reason (bool, optional): Flag to include a reason for the plan. Defaults to False.
        num_step (int, optional): Number of steps in the plan. Defaults to 3.
        **kwargs: Additional keyword arguments for further customization.

    Returns:
        Any: The result of the planning task.
    """

    branch = branch or Branch()
    unit = Unit(branch)

    return await unit.plan(
        instruction=instruction,
        context=context,
        system=system,
        sender=sender,
        recipient=recipient,
        form=form,
        confidence_score=confidence_score,
        reason=reason,
        num_step=num_step,
        **kwargs,
    )
